import tzlocal
import warnings
import mpmath
import pandas as pd
from datetime import timedelta

class Time(object):
    """Time class encapsulates provides a set of methods for conversion between various time formats.
    In addition it provides a generator for days between given periods"""

    __units = {'ns': 1, 'us': 1e3, 'ms': 1e6, 's': 1e9}

    def __init__(self, input, unit=None, tz=None):

        if isinstance(input, (int, float)) and unit in Time.__units.keys():
            input = Time.__convertRealNumbersToUnixTime(input, Time.__units[unit])
            unit = 'ns'

        if tz:
            self.__pdTimestamp = pd.Timestamp(input, tz=tz, unit=unit)
        else:
            self.__pdTimestamp = pd.Timestamp(input, tz=tzlocal.get_localzone(), unit=unit)

        # region properties

    @property
    def unixTimestamp(self):
        return self.__pdTimestamp.value

    @property
    def pandasTimestamp(self):
        return self.__pdTimestamp

    @property
    def stringTimestamp(self):
        return self.__pdTimestamp.__str__()

    @property
    def stringTimestampWithoutTimeZoneIndication(self):
        return self.__pdTimestamp.__str__()[:-6]

    @property
    def datetimeTimestamp(self):
        return self.__pdTimestamp.to_pydatetime()

    # endregion

    # region conversion methods

    @staticmethod
    def convertToUnixTimestamp(input, unit='ns', tz=None):
        """
            **Converts input time variable into a unix timestamp integer type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into unix timestamp integer.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  stringDate = "2018-05-01 00:00:00"
                  Time.convertToUnixTimestamp(stringDate, unit = "ns")
                  >>> 1525125600000000000

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  dateTime = datetime(2018, 5, 1, 0, 0, 0, 0)
                  Time.convertToUnixTimestamp(stringDate, unit = "ns")
                  >>> 1525125600000000000

                  from lhcsmapi.Time import Time
                  timestamp = 1525125600000000000
                  Time.convertToUnixTimestamp(stringDate, unit = "ns")
                  >>> 1525125600000000000

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pandasTimestamp = pd.Timestamp(2018, 5, 1, 0)
                  Time.convertToUnixTimestamp(pandasTimestamp, unit = "ns")
                  >>> 1525125600000000000

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  stringDate = "2018-05-01 00:00:00"
                  Time.convertToUnixTimestamp(stringDate, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        timeConverter = Time.__validateConversion(input, 'unix', unit, tz)
        return timeConverter.unixTimestamp

    @staticmethod
    def convertToPandasTimestamp(input, unit='ns', tz=None):
        """
            **Converts input time variable into a pandas.Timestamp type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into pandas.Timestamp.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  stringDate = "2019-07-10 10:23:27.013"
                  Time.convertToPandasTimestamp(stringDate, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  dateTime = datetime(2019, 7, 10, 10, 23, 27, 13)
                  Time.convertToPandasTimestamp(stringDate, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

                  from lhcsmapi.Time import Time
                  timestamp = 1562747007013000000
                  Time.convertToPandasTimestamp(timestamp, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pandasTimestamp = pd.Timestamp(2019, 7, 10, 10, 23, 27, 13000)
                  Time.convertToPandasTimestamp(pandasTimestamp, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  stringDate = "2019-07-10 10:23:27.013"
                  Time.convertToPandasTimestamp(stringDate, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        timeConverter = Time.__validateConversion(input, 'pandas', unit, tz)
        return timeConverter.pandasTimestamp

    @staticmethod
    def convertToString(input, unit='ns', tz=None):
        """
            **Converts input time variable into a string type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into string.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  stringDate = "2017-07-02 01:23:27"
                  Time.convertToString(stringDate, unit = "ns")
                  >>> '2017-07-02 01:23:27'

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  dateTime = datetime(2017, 7, 2, 1, 23, 27, 0)
                  Time.convertToString(stringDate, unit = "ns")
                  >>> '2017-07-02 01:23:27'

                  from lhcsmapi.Time import Time
                  timestamp = 1498951407000000000
                  Time.convertToString(timestamp, unit = "ns")
                  >>> '2017-07-02 01:23:27'

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pandasTimestamp = pd.Timestamp(2017, 7, 2, 1, 23, 27)
                  Time.convertToString(pandasTimestamp, unit = "ns")
                  >>> '2017-07-02 01:23:27'

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  stringDate = "2017-07-02 01:23:27"
                  Time.convertToString(stringDate, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        timeConverter = Time.__validateConversion(input, 'string formatted', unit, tz)
        return timeConverter.stringTimestamp

    @staticmethod
    def convertToDatetime(input, unit='ns', tz=None):
        """
            **Converts input time variable into a datetime type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into datetime.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  stringDate = "2019-01-02 23:59:11"
                  Time.convertToDatetime(stringDate, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  dateTime = datetime(2019, 1, 2, 23, 59, 11, 0)
                  Time.convertToDatetime(stringDate, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

                  from lhcsmapi.Time import Time
                  timestamp = 1546469951000000000
                  Time.convertToDatetime(timestamp, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pandasTimestamp = pd.Timestamp(2019, 1, 2, 23, 59, 11, 0)
                  Time.convertToDatetime(pandasTimestamp, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  stringDate = "2019-01-02 23:59:11"
                  Time.convertToDatetime(stringDate, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        timeConverter = Time.__validateConversion(input, 'datetime', unit, tz)
        return timeConverter.datetimeTimestamp

    @staticmethod
    def convertToUnixTimeSeconds(input, unit='ns', tz=None, warn=False):
        """
            **Converts input time variable into a float type with us precision**

            Method returns input converted to unix time in seconds with us precision

            :param input: input unix timestamp in nanoseconds
            :type input: int unix timestamp
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param durationData: time zone (optional, local by default)
            :type durationData:
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::



            - Expected Fail Response::


        """

        unixTime = Time.convertToUnixTimestamp(input, unit, tz)
        return Time.__convertUnixTimeInNanosecondsToSeconds(unixTime, warn)

    # endregion
    @staticmethod
    def getQueryPeriodInUnixTime(startTimeData=None, endTimeData=None, durationData=None, tz=None):
        """
            **Calculates start and end time from start time and either end time or duration**

            Method calculates startTime and endTime or duration values based on the arguments and then returns a tuple with these
            converted to unix time. Three operation modes are possible:

            - if endTimeData is not None:

            (startTime, endTime) = (startTimeData, endTimeData)

            - elif len(durationData) is 1:

            (startTime, endTime) = (startTimeData, startTimeData+durationData[0])

            - elif len(durationData) is 2:

            (startTime, endTime) = (startTimeData-durationData[0], startTimeData+durationData[1])

            :param startTimeData: input date and time
            :type startTimeData: int unix timestamp
            :param endTimeData: unit of the input (None if durationData defined)
            :type endTimeData: int unix timestamp
            :param durationData: time zone (None if endTimeData defined)
            :type durationData: list of (time, unit) tuples
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  startTimeData = 1426220459491000000
                  endTimeData =   1426220479491000000
                  Time.getQueryPeriodInUnixTime(startTimeData, endTimeData, None, tz='Europe/Zurich')
                  >>> (1426220459491000000, 1426220479491000000)

                  from lhcsmapi.Time import Time
                  startTimeData = 1426220459491000000
                  durationData =  [(20, 's')]
                  Time.getQueryPeriodInUnixTime(startTimeData, None, durationData, tz='Europe/Zurich')
                  >>> (1426220459491000000, 1426220479491000000)

                  from lhcsmapi.Time import Time
                  startTimeData = 1426220469491000000
                  durationData =  [(10, 's'), (10, 's')]
                  Time.getQueryPeriodInUnixTime(startTimeData, None, durationData, tz='Europe/Zurich')
                  >>> (1426220459491000000, 1426220479491000000)

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  startTimeData = 1426220469491000000
                  durationData =  [(10, 's'), (10, 's')]
                  Time.getQueryPeriodInUnixTime(startTimeData, None, None, tz='Europe/Zurich')
                  >>> ValueError: endTimeData and durationData can not be both None at the same time!
        """

        #if not any([endTimeData, durationData]):
        #    raise ValueError("endTimeData and durationData can not be both None at the same time!")

        # convert startTime to unixTime:
        startTime, unit = Time.__validateTimeTuple(startTimeData)
        if startTime:
            startTime = Time.convertToUnixTimestamp(startTime, unit, tz)

        # extact endTime if provided if so convert to unixTime:
        endTime, unit = Time.__validateTimeTuple(endTimeData)
        if endTime:
            endTime = Time.convertToUnixTimestamp(endTime, unit, tz)

        elif durationData:

            if isinstance(durationData, tuple):
                durationData = [durationData]

            # duration is list with single entry or tuple , calculate endTime from duration:
            if isinstance(durationData, list) and len(durationData) == 1:

                duration, unit = Time.__validateTimeTuple(durationData[0])
                endTime = startTime + Time.convertToUnixTimestamp(duration, unit, tz)

            # duration is list with two entries (tuple) , calculate startTime and endTime from duration:
            elif isinstance(durationData, list) and len(durationData) == 2:

                dt1, unit1 = Time.__validateTimeTuple(durationData[0])
                dt2, unit2 = Time.__validateTimeTuple(durationData[1])

                endTime = startTime + Time.convertToUnixTimestamp(dt2, unit2, tz)
                startTime -= Time.convertToUnixTimestamp(dt1, unit1, tz)

        return (startTime, endTime)

    @staticmethod
    def daterange(startDate, endDate):
        """
            **Creates a generator of consecutive days between two dates**

            Method allowing users to create a generator of days from a given
            startDate, with an increment of a day, to endDay - 1 day

            :param startDate: starting date for the generator
            :type input: datetime.date(YYYY, MM, DD)
            :param endDate: end date
            :type input: datetime.date(YYYY, MM, DD)
            :return: generator providing a day from startDate to endDate - 1 day

            - Example::

                  from lhcsmapi.Time import Time
                  from datetime import date
                  startDate = date(2014, 1, 1)
                  endDate = date(2014, 1, 3)
                  for singleDate in Time.daterange(startDate, endDate):
                      print(singleDate)
                  >>> 2014-01-01
                  >>> 2014-01-02

            - Expected Fail Response::

                  from datetime import date
                  startDate = date(2014, 1, 5)
                  endDate = date(2014, 1, 3)
                  for singleDate in Time.daterange(startDate, endDate):
                      print(singleDate)
                  >>>
        """
        for n in range(int((endDate - startDate).days)):
            yield startDate + timedelta(n)

    @staticmethod
    def moduloDayInUnixTimestamp(startTime, endTime):
        startTimeUnix = Time.convertToUnixTimestamp(startTime)
        endTimeUnix = Time.convertToUnixTimestamp(endTime)
        if startTimeUnix > endTimeUnix:
            raise AttributeError('startTime {} is greater than the endTime {}'.format(startTime, endTime))

        diffTimeUnix = endTimeUnix - startTimeUnix
        dayInNanos = int(24*60*60*1e9)

        quotient = diffTimeUnix//dayInNanos
        remainder = diffTimeUnix%dayInNanos
        return (quotient, remainder)


    # region private helper methods
    @staticmethod
    def __validateTimeTuple(timeData):
        """Method returns a tuple that can be safely unpacked regardless of whether timedata is a  tuple: (time, unit) or
        single: time object"""

        time = None
        unit = None

        if isinstance(timeData, tuple):
            if len(timeData) > 0:
                time = timeData[0]

            if len(timeData) > 1:
                unit = timeData[1]

        else:
            time = timeData

        return (time, unit)

    @staticmethod
    def __validateConversion(input, conversionType, unit='ns', tz=None):
        try:
            return Time(input, unit, tz)
        except TypeError:
            raise TypeError('Cannot convert input {} to {} timestamp'.format(type(input), conversionType))

    @staticmethod
    def __convertRealNumbersToUnixTime(input, ratio):
        """Method converts input of int or float to unix time based on ratio"""
        with mpmath.mp.workdps(20):
            inputFull = mpmath.mpf(str(input))

            ratio = mpmath.mpf(ratio)
            unixTime = inputFull * ratio
            return int(unixTime)

    @staticmethod
    def __convertUnixTimeInNanosecondsToSeconds(unixTime, warn=False):
        """Method converts unix time in nanoseconds to unix time in seconds with us precision"""

        nsRemainder = unixTime % 1000

        if warn and nsRemainder > 0:
            warnings.warn('discarding nonzero {} nanoseconds in conversion'.format(nsRemainder), stacklevel=2)

        ns = unixTime - nsRemainder
        unixTimeSeconds = mpmath.mpf(ns) / mpmath.mpf(1e9)

        return float(unixTimeSeconds)
    #endregion
