import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ..metadata.SignalMetadata import SignalMetadata
from ..metadata.MappingMetadata import MappingMetadata
from ..Time import Time
from ..dbsignal.Signal import Signal
from ..post_mortem.PmDbRequest import PmDbRequest
from ..dbsignal.SignalAnalysis import SignalAnalysis
from ..dbsignal.SignalUtilities import SignalUtilities

class DiodeLeadResistanceAnalysis(object):

    @staticmethod
    def get_magnet_type_i_meas_u_diode(magnet, circuit_name):
        circuit_type = circuit_name.split('.')[0]
        magnet_type = "M" + circuit_type[1:2]

        i_meas = SignalMetadata.getVariableName(circuit_type, circuit_name, systemType="PC", databaseName="CALS",
                                                variableName="I_MEAS")
        u_diode = SignalMetadata.getVariableName(circuit_type, circuit_name, systemType="DIODE_{}".format(circuit_type),
                                                 databaseName='CALS', variableName="U_DIODE_{}".format(circuit_type))
        u_diode = u_diode.replace("%MAGNET%", "{}.{}".format(magnet_type, magnet))

        return magnet_type, i_meas, u_diode

    @staticmethod
    def set_to_zero_resistance_outside_of_range(u_diode, I_MEAS):
        calc_diff_u_diode = lambda x: x[-1] - x[0]
        u_diode['dU_DIODE'] = u_diode['U_DIODE_min'].rolling(11).apply(calc_diff_u_diode)

        u_diode.loc[u_diode.index < u_diode.index[530], 'dU_DIODE'] = 0

        if u_diode['dU_DIODE'].max() > 0.05:
            index_cut = u_diode.query('dU_DIODE > 0.05').index[0]
        else:
            index_cut = max(u_diode.index)

        u_diode.loc[u_diode[I_MEAS] <= 100, 'RES'] = 0
        u_diode.loc[u_diode.index > index_cut, 'RES'] = 0

    @staticmethod
    def calculate_minimum_diode_and_current_leads_voltage(u_diode, U_DIODE):
        # "5" - toggling period of nQPS U_DIODE signals
        # "50s x 10 Hz + 20" delay to the diode full opening
        calc_u_diode_min = lambda x: min((x[0] + x[-1]) / 2, x[5])
        u_diode['U_DIODE_min'] = u_diode.loc[u_diode.index >= u_diode.index[515], U_DIODE].rolling(11).apply(
            calc_u_diode_min).shift(-5)

        calc_u_res = lambda x: abs(x[5] - (x[0] + x[-1]) / 2)
        u_diode['U_RES'] = u_diode.loc[u_diode.index >= u_diode.index[515], U_DIODE].rolling(11).apply(calc_u_res).shift(-5)

        u_diode.loc[u_diode.index < u_diode.index[520], ['U_RES', 'U_DIODE_min']] = 0

    @staticmethod
    def get_cals_signal(signal_name, timestamp, magnet_type, *, ldb=None):
        # get start and end time
        timestamp_unix_sec = Time.convertToUnixTimeSeconds(timestamp)
        t_end = 350 if magnet_type == "MB" else 150

        # query signal
        signal = Signal().read('CALS', name=signal_name, startTime=(timestamp_unix_sec, 's'),
                               duration=[(50, 's'), (t_end, 's')], ldb=ldb)
        return SignalUtilities.synchronizeDataframe(signal, timestamp_unix_sec)

    @staticmethod
    def filter_resistance(u_diode):
        u_diode['RES'] = u_diode['RES'].rolling(3).median().values
        u_diode['RES_filtered'] = u_diode['RES'].rolling(window=20).mean()

    @staticmethod
    def extract_res_max_time_current_cals(u_diode, I_MEAS):
        res_max = u_diode['RES_filtered'].max()
        t_res_max = u_diode.idxmax(axis=0, skipna=True)['RES_filtered']
        i_res_max = u_diode.loc[u_diode.index == t_res_max, I_MEAS].values[0]
        return res_max, t_res_max, i_res_max

    @staticmethod
    def calculate_resistance_cals(magnet, circuit_name, timestamp, ldb=None):
        # get metadata - signal names
        magnet_type, I_MEAS, U_DIODE = DiodeLeadResistanceAnalysis.get_magnet_type_i_meas_u_diode(magnet, circuit_name)

        # query circuit current
        i_circuit_sync = DiodeLeadResistanceAnalysis.get_cals_signal(I_MEAS, timestamp, magnet_type, ldb=ldb)
        t_end = 350 if magnet_type == "MB" else 150
        i_circuit_sync_resampled = SignalAnalysis.resampleSeries(i_circuit_sync, ts=0.1, t0=-50, t_end=t_end)

        # query diode voltage
        u_diode_sync = DiodeLeadResistanceAnalysis.get_cals_signal(U_DIODE, timestamp, magnet_type, ldb=ldb)

        # join current and diode voltage into a single dataframe
        u_diode_sync[I_MEAS] = i_circuit_sync_resampled[I_MEAS].values

        # recover resistive and diode minimum voltage
        DiodeLeadResistanceAnalysis.calculate_minimum_diode_and_current_leads_voltage(u_diode_sync, U_DIODE)

        # calculate resistance
        u_diode_sync['RES'] = u_diode_sync.apply(lambda row: 1e+6 * row['U_RES'] / row[I_MEAS], axis=1, raw=False)

        # set to 0 resistance values for too low current and with too high diode voltage derivative
        DiodeLeadResistanceAnalysis.set_to_zero_resistance_outside_of_range(u_diode_sync, I_MEAS)

        # filter resistance
        DiodeLeadResistanceAnalysis.filter_resistance(u_diode_sync)

        return u_diode_sync

    @staticmethod
    def plot_resistance_cals(magnet, circuit_name, u_diode):
        magnet_type, I_MEAS, U_DIODE = DiodeLeadResistanceAnalysis.get_magnet_type_i_meas_u_diode(magnet, circuit_name)
        res_max_cals, t_res_max_cals, i_res_max_cals = DiodeLeadResistanceAnalysis.extract_res_max_time_current_cals(
            u_diode, I_MEAS)

        ax1 = u_diode[I_MEAS].plot(figsize=(15, 7.5), legend=None, lw=2, color='C3')
        ax1.set_xlabel("time, [s]", fontsize=20)
        ax1.tick_params(axis='both', which='major', labelsize=15)
        ax1.set_ylabel("I_MEAS [A]", color='C3', fontsize=20)
        ax1.grid()
        ax2 = ax1.twinx()
        u_diode[U_DIODE].plot(ax=ax2, legend=None, lw=2, color='C0')
        ax2.set_ylabel("U_DIODE [V]", color='C0', fontsize=20)
        ax2.tick_params(axis='both', which='major', labelsize=15)
        plt.show()

        #
        ax1 = u_diode[I_MEAS].plot(title='I_MEAS(t) and diodeLeadResistance(t)', figsize=(15, 7.5), legend=None, lw=2,
                                   color='C3')
        ax1.title.set_size(20)
        ax1.set_xlabel("time, [s]", fontsize=20)
        ax1.tick_params(axis='both', which='major', labelsize=15)
        ax1.set_ylabel("I_MEAS [A]", color='C3', fontsize=20)
        ax1.grid()
        ax2 = ax1.twinx()
        u_diode['RES_filtered'].plot(ax=ax2, legend=None, lw=2, color='C0')
        ax2.set_ylabel("Resistance [uOhm]", color='C0', fontsize=20)
        ax2.tick_params(axis='both', which='major', labelsize=15)

        x_text = 200 if magnet_type == "MB" else 100
        plotText = ('r_max = ' + "{:.1f}".format(res_max_cals) + ' uOhm' + '\n'
                    + 'i_max = ' + str(int(i_res_max_cals)) + ' A' + '\n'
                    + 't_max = ' + str(int(t_res_max_cals)) + ' s')
        ax2.annotate(plotText, xytext=(x_text, res_max_cals / 2), xy=(t_res_max_cals, res_max_cals),
                     arrowprops=dict(facecolor='white', shrink=0.05), fontsize=20)
        plt.show()

    @staticmethod
    def find_diode_pm_entries(circuit_type, circuit_name, magnet, timestamp_iqps_qh, duration=[(120, 's'), (600, 's')]):
        # 120 s <- exception to be discussed with QPS gurus

        system_diode = 'DIODE_{}'.format(circuit_type)
        metadata_diode = SignalMetadata.getCircuitVariableDatabaseMetadata(circuit_type, circuit_name, system_diode,
                                                                           'PM')
        class_name_qps, system_qps = metadata_diode['className'], metadata_diode['system']
        source_qps = MappingMetadata.getCrateNameFromMagnetName(circuit_type, magnet)

        event_source_timestamp_nqps = PmDbRequest.findEvents(source_qps, system_qps, class_name_qps,
                                                             startTime=timestamp_iqps_qh, duration=duration)

        return [timestamp for _, timestamp in event_source_timestamp_nqps]

    @staticmethod
    def find_pc_pm_entries(circuit_type, circuit_name, timestamp_iqps_qh, duration=[(100, 's'), (1, 's')]):
        metadata_fgc = SignalMetadata.getCircuitVariableDatabaseMetadata(circuit_type, circuit_name, 'PC', 'PM')
        class_name_fgc, source_fgc, system_fgc = metadata_fgc['className'], metadata_fgc['source'], metadata_fgc[
            'system']

        event_source_timestamp_pc = PmDbRequest.findEvents(source_fgc, system_fgc, class_name_fgc,
                                                           startTime=timestamp_iqps_qh, duration=duration)

        return [timestamp for _, timestamp in event_source_timestamp_pc]

    @staticmethod
    def preprocess_pm_qps_signal(time_corr, timestamp_unix, pm_qps_signal, pm_signal_name):
        pm_qps_signal.drop(pm_qps_signal.index[:2], inplace=True)
        pm_qps_signal.drop(pm_qps_signal.index[-2:], inplace=True)
        pm_qps_signal.drop(pm_qps_signal[pm_qps_signal[pm_signal_name] == 0.0].index, inplace=True)
        pm_qps_signal = pm_qps_signal.set_index(((pm_qps_signal.index.values) - timestamp_unix) / 1e+9 + time_corr)
        return pm_qps_signal

    @staticmethod
    def preprocess_pm_pc_signal(iab_i_a, iab_i_a_name, index_end_diode, index_start_diode, timestamp_unix):
        iab_i_a.drop(iab_i_a[iab_i_a[iab_i_a_name] == 0.0].index, inplace=True)
        iab_i_a = iab_i_a.set_index(((iab_i_a.index.values) - timestamp_unix) / 1e+9)

        index_end = iab_i_a.index.get_loc(index_end_diode, method='nearest')
        index_start = iab_i_a.index.get_loc(index_start_diode, method='nearest')
        iab_i_a.drop(iab_i_a.index[index_end:], inplace=True)
        iab_i_a.drop(iab_i_a.index[:index_start], inplace=True)
        return iab_i_a

    @staticmethod
    def resample_pm_signal(df):
        df_resampled = df.reindex(
            df.index.union(np.linspace(df.index[0], df.index[-1], 2001))).interpolate('index')
        df_resampled.drop(df.index, axis=0, inplace=True)
        return df_resampled

    @staticmethod
    def calculate_resistence(t_max, df):
        df['U_RES'] = df['U_DIODE_A'] - df['U_DIODE_B']
        df.loc[df.index < t_max + 0.05, 'U_RES'] = 0
        df['RES'] = df['U_RES'] / df['IAB.I_A']
        df['RES'] = df['RES'].rolling(window=3, min_periods=1).median()

    @staticmethod
    def extract_res_max_time_current_pm(df):
        res_max = int(df['RES'].max() * 1e+6)
        t_res_max = df.idxmax(axis=0, skipna=True)['RES']
        i_res_max = df.reindex([t_res_max], method='nearest', tolerance=.0001)['IAB.I_A'].values[0]
        return res_max, t_res_max, i_res_max

    @staticmethod
    def calculate_resistance_pm(circuit_type, circuit_name, magnet, timestamp_iqps_qh, timestamp_nqps,
                                timestamp_fgc, time_corr=0.0):
        # get source qps
        source_qps = MappingMetadata.getCrateNameFromMagnetName(circuit_type, magnet)

        # get timestamps for nqps and fgc

        # get metadata
        magnet_type, i_meas_name, u_diode_name = DiodeLeadResistanceAnalysis.get_magnet_type_i_meas_u_diode(magnet, circuit_name)

        # get timestamp
        timestamp_unix = Time.convertToUnixTimestamp(timestamp_iqps_qh)

        # query U_DIODE
        system_diode = 'DIODE_RB' if magnet_type == "MB" else 'DIODE_RQD'
        metadata_diode = SignalMetadata.getCircuitVariableDatabaseMetadata(circuit_type, circuit_name, system_diode, 'PM')
        class_name_qps, system_qps = metadata_diode['className'], metadata_diode['system']
        # Board A
        u_diode_a = Signal().read('pm', system=system_qps, className=class_name_qps, source=source_qps,
                                  signalName=u_diode_name, eventTime=timestamp_nqps[0])
        u_diode_a = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_diode_a,
                                                                         u_diode_name)
        # Board B
        u_diode_b = Signal().read('pm', system=system_qps, className=class_name_qps, source=source_qps,
                                  signalName=u_diode_name, eventTime=timestamp_nqps[1])
        u_diode_b = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix - 1e+6, u_diode_b,
                                                                         u_diode_name)

        # U_REF
        u_ref_name = SignalMetadata.getVariableName(circuit_type, circuit_name, system_diode, 'PM', 'U_REF_N1').replace(
            '%QPS_CRATE%', source_qps)
        u_ref = Signal().read('pm', system='QPS', className=class_name_qps, source=source_qps, signalName=u_ref_name,
                              eventTime=timestamp_nqps[0])
        u_ref = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_ref, u_ref_name)

        # I_A
        metadata_fgc = SignalMetadata.getCircuitVariableDatabaseMetadata(circuit_type, circuit_name, 'PC', 'PM')
        class_name_fgc, source_fgc = metadata_fgc['className'], metadata_fgc['source']
        iab_i_a_name = SignalMetadata.getVariableName(circuit_type, circuit_name, 'PC', 'PM', 'I_A')

        iab_i_a = Signal().read('pm', system='FGC', className=class_name_fgc, source=source_fgc,
                                signalName=iab_i_a_name, eventTime=timestamp_fgc[0])

        index_start_diode = u_diode_a.index[0]
        index_end_diode = u_diode_a.index[-1]
        iab_i_a = DiodeLeadResistanceAnalysis.preprocess_pm_pc_signal(iab_i_a, iab_i_a_name, index_end_diode,
                                                                      index_start_diode, timestamp_unix)

        # resample signals
        u_diode_a_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_diode_a)
        u_diode_b_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_diode_b)
        u_ref_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_ref)
        iab_i_a_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(iab_i_a)

        t_max = u_diode_a.idxmax(axis=0, skipna=True)[0]

        # Calculate resistance
        res = pd.DataFrame()
        res['U_DIODE_A'] = u_diode_a_resampled[u_diode_name]
        res['U_DIODE_B'] = u_diode_b_resampled.values
        res['U_REF'] = u_ref_resampled.values
        res['IAB.I_A'] = iab_i_a_resampled.values

        DiodeLeadResistanceAnalysis.calculate_resistence(t_max, res)

        return res

    @staticmethod
    def plot_resistance_pm(circuit_name, res_df):
        res_max_pm, t_res_max_pm, i_res_max_pm = DiodeLeadResistanceAnalysis.extract_res_max_time_current_pm(res_df)

        ax1 = res_df['IAB.I_A'].plot(figsize=(15, 10), legend=None, lw=2, xlim=(-0.2, 1.0), color='C3')
        ax1.set_xlabel("time, [s]", fontsize=20)
        ax1.tick_params(axis='both', which='major', labelsize=15)
        ax1.set_ylabel("I_MEAS [A]", color='C3', fontsize=20)
        ax1.grid()
        ax2 = ax1.twinx()
        res_df['U_DIODE_A'].plot(ax=ax2, legend=None, lw=2, xlim=(-0.2, 1.0), color='C0')
        res_df['U_DIODE_B'].plot(ax=ax2, legend=None, lw=2, xlim=(-0.2, 1.0), color='C0')
        res_df['U_REF'].plot(ax=ax2, legend=None, lw=2, xlim=(-0.2, 1.0), color='C2')
        ax2.set_ylabel("U_DIODE [V]", color='C0', fontsize=20)
        ax2.tick_params(axis='both', which='major', labelsize=15)
        plt.show()
        x_text = 0.5 if "RB" in circuit_name else 1
        #
        plot_text = ('res_max = ' + "{:.1f}".format(res_max_pm) + ' uOhm' + '\n'
                     + 'i_res_max = ' + str(int(i_res_max_pm)) + ' A' + '\n'
                     + 't_res_max = ' + "{:.3f}".format(t_res_max_pm) + ' s')
        #
        ax = res_df['RES'].plot(title='Diode Lead Resistance [Ohm] & Current [A]', figsize=(15, 7.6), legend='True',
                                grid='True')
        res_df['IAB.I_A'].plot(ax=ax, secondary_y=True, legend='True', grid='True')
        ax.annotate(
            plot_text, xytext=(x_text, res_max_pm * 1e-6 / 2), xy=(t_res_max_pm, res_max_pm * 1e-6),
            arrowprops=dict(facecolor='white', shrink=0.05), fontsize=15)