import numpy as np
import pandas as pd
from copy import deepcopy
from functools import reduce
from scipy import signal as s
import matplotlib.pyplot as plt

from ..dbsignal.Signal import Signal
from ..dbsignal.SignalElectrical import SignalElectrical
from ..dbsignal.SignalUtilities import SignalUtilities
from ..dbsignal.ExponentialSignalAnalysis import ExponentialSignalAnalysis
from ..dbsignal.SignalAnalysis import SignalAnalysis
from ..dbsignal.SignalProcessing import SignalProcessing
from ..metadata.SignalMetadata import SignalMetadata

class QuenchHeaterAnalysis(object):

    @staticmethod
    def query_qh_discharge(circuit_type, circuit_name, source_qh, timestamp_qh):
        # Metadata
        U_HDS = SignalMetadata.getVariableName(circuit_type, circuit_name, 'QH', 'PM', 'U_HDS')
        I_HDS = SignalMetadata.getVariableName(circuit_type, circuit_name, 'QH', 'PM', 'I_HDS')

        metadata_qh = SignalMetadata.getCircuitVariableDatabaseMetadata(circuit_type, circuit_name, 'QH', 'PM')
        system_qh, classname_qh = metadata_qh['system'], metadata_qh['className']

        # query post-mortem
        qh_signal = Signal()
        for qh_signal_name in U_HDS + I_HDS:
            qh_signal.read('pm', append=True, system=system_qh, className=classname_qh, source=source_qh,
                           signalName=qh_signal_name, eventTime=timestamp_qh)

        return qh_signal.df

    @staticmethod
    def preprocess_qh_discharge(circuit_type, circuit_name, qh_df, timestamp_qh):
        # Metadata
        U_HDS = SignalMetadata.getVariableName(circuit_type, circuit_name, 'QH', 'PM', 'U_HDS')
        I_HDS = SignalMetadata.getVariableName(circuit_type, circuit_name, 'QH', 'PM', 'I_HDS')

        # synchronize time to 0
        qh_sync_df = SignalUtilities.synchronizeDataframe(qh_df, timestamp_qh)
        # convert indices to seconds
        qh_sync_df = SignalUtilities.convertIndicesToSeconds(qh_sync_df)
        # extract decay only
        qh_sync_decay_df = ExponentialSignalAnalysis.getDfwithDecayOnly(qh_sync_df)
        # filter
        qh_sync_decay_df = qh_sync_decay_df.rolling(window=3, min_periods=1).median()
        # calculate resistance
        if circuit_type == "RB":
            R_HDS = [u.replace('U', 'R') for u in U_HDS]
            qh_sync_decay_df = SignalElectrical.calculateResistances(qh_sync_decay_df, U_HDS, I_HDS, R_HDS)

        return qh_sync_decay_df

    @staticmethod
    def plot_qh_discharge(circuit_type, qh_sync_decay_df, qh_ref_sync_decay_df):
        no_subplots = 3 if circuit_type == "RB" else 1

        fig, ax = plt.subplots(1, no_subplots, figsize=(20, 10))
        qh_sync_decay_df.filter(regex=("U_HDS_\d{1}")).plot(ax=ax[0]).set(xlabel="[s]", ylabel="[V]")
        qh_ref_sync_decay_df.filter(regex=("U_HDS_\d{1}")).plot(ax=ax[0], style='--').set(xlabel="[s]", ylabel="[V]")
        ax[0].grid(True)

        if circuit_type == "RB":
            qh_sync_decay_df.filter(regex=("I_HDS_\d{1}")).plot(ax=ax[1]).set(xlabel="[s]", ylabel="[A]")
            qh_ref_sync_decay_df.filter(regex=("I_HDS_\d{1}")).plot(ax=ax[1], style='--').set(xlabel="[s]", ylabel="[A]")
            ax[1].grid(True)

            qh_sync_decay_df.filter(regex=("R_HDS_\d{1}")).plot(ax=ax[2]).set(xlabel="[s]", ylabel="[A]")
            qh_ref_sync_decay_df.filter(regex=("R_HDS_\d{1}")).plot(ax=ax[2], style='--').set(xlabel="[s]", ylabel="[A]")
            ax[2].grid(True)

    @staticmethod
    def calculate_qh_discharge_features(circuit_type, circuit_name, qh_sync_decay_df, timestamp_qh):
        # Metadata
        U_HDS = SignalMetadata.getVariableName(circuit_type, circuit_name, 'QH', 'PM', 'U_HDS')
        I_HDS = SignalMetadata.getVariableName(circuit_type, circuit_name, 'QH', 'PM', 'I_HDS')

        char_time_qh = QuenchHeaterAnalysis.calculateCharacteristicTimeConstant(qh_sync_decay_df, timestamp_qh,
                                                                                U_HDS + I_HDS,
                                                                                methods=['chargeApproach'])
        char_time_qh = char_time_qh.filter(regex='[UI]_HDS_\d{1}_chargeApproach')
        char_time_qh.columns = [col.replace('chargeApproach', 'tau') for col in char_time_qh.columns]
        if circuit_type is 'RB':
            R_HDS = [u.replace('U', 'R') for u in U_HDS]
            u_i_initial_final_qh = QuenchHeaterAnalysis.calculateInitialFinalValue(qh_sync_decay_df, timestamp_qh,
                                                                                   U_HDS + I_HDS)
            r_initial_qh = QuenchHeaterAnalysis.calculateInitialResistance(u_i_initial_final_qh, timestamp_qh, U_HDS,
                                                                           I_HDS, R_HDS)
            features_qh = char_time_qh.join(r_initial_qh)
        else:
            features_qh = char_time_qh
        return features_qh

    @staticmethod
    def rate_qh_discharge(circuit_type, features_qh_concat):
        features_abs_diff_qh = features_qh_concat.diff().tail(1).abs()
        # reduce operation
        is_u_hds_tau_correct = reduce(np.logical_and,
                                      [features_abs_diff_qh['U_HDS_{}_tau'.format(i)] < 0.003 for i in range(1, 4)])
        if circuit_type == 'RB':
            is_i_hds_tau_correct = reduce(np.logical_and,
                                          [features_abs_diff_qh['I_HDS_{}_tau'.format(i)] < 0.003 for i in range(1, 4)])
            is_r_hds_initial_correct = reduce(np.logical_and,
                                              [features_abs_diff_qh['R_HDS_{}_initial'.format(i)] < 0.5 for i in
                                               range(1, 4)])
            return is_u_hds_tau_correct.values[0] and is_i_hds_tau_correct.values[0] and \
                   is_r_hds_initial_correct.values[0]
        else:
            return is_u_hds_tau_correct.values[0]

    @staticmethod
    def calculateInitialFinalValue(df, timestamp, signalNames):
        rowFeatureDf = pd.DataFrame(index=[timestamp])
        for signalName in signalNames:
            rowFeatureDf[signalName + '_initial'] = np.mean(s.medfilt(df[signalName].values[0:19], 3))
            rowFeatureDf[signalName + '_final'] = np.mean(s.medfilt(df[signalName].values[-20:-1], 3))

        return rowFeatureDf

    @staticmethod
    def calculateInitialResistance(rowFeatureDfInput, timestamp, uNames, iNames, rNames):
        rowFeatureDf = pd.DataFrame(index=[timestamp])
        for u, i, r in zip(uNames, iNames, rNames):
            rowFeatureDf[r + '_initial'] = rowFeatureDfInput[u + '_initial'] / rowFeatureDfInput[i + '_initial']
        return rowFeatureDf

    # ToDo: for the sake of flexibility, everything should be calculated as a method, including cTime_mean and std
    @staticmethod
    def calculateCharacteristicTimeConstant(df, timestamp, signalNames,
                                            methods=["chargeApproach", "energyApproach", "linReg", "expFit"]):
        rowFeatureDf = pd.DataFrame(index=[timestamp])
        for signalName in signalNames:
            cTimeOfExpDecay = ExponentialSignalAnalysis.getCharacteristicTimeOfExponentialDecay(
                df[signalName].to_frame(), methods)
            rowFeatureDfTau = SignalProcessing.convertFeatureDataframeIntoRow(cTimeOfExpDecay, timestamp)
            rowFeatureDf = pd.concat([rowFeatureDf, rowFeatureDfTau], axis=1, sort=False)
            tauDf = ExponentialSignalAnalysis.calculateChangeInCharacteristicOfExponentialDecay(df[signalName],
                                                                                                window=1000)
            rowFeatureDf[signalName + '_cTime_mean'] = tauDf.mean()
            rowFeatureDf[signalName + '_cTime_std'] = tauDf.std()

        return rowFeatureDf

    @staticmethod
    def calculateSimilarityMatrix(df, timestamp, signalNames):
        rowFeatureDf = pd.DataFrame(index=[timestamp])
        for signals in signalNames:
            similarityMatrix = SignalAnalysis.calculateSimilarityMatrix(df[signals]).fillna(-1)
            rowFeatureDfSmCurrent = SignalProcessing.convertFeatureDataframeIntoRow(similarityMatrix, timestamp)
            rowFeatureDf = pd.concat([rowFeatureDf, rowFeatureDfSmCurrent], axis=1, sort=False)

        return rowFeatureDf

    @staticmethod
    def performCurveToCurveComparisonGroup(myDFDecay, myDFDecayRef, rowFeatureDf, signalNames, envelopeAmplitude,
                                           useGlobalEnvelopeDecayTimeConstant=False):
        # Filter Data
        myDFfiltered = myDFDecay.rolling(window=50, min_periods=1).median()
        myDFfilteredRef = myDFDecayRef.rolling(window=50, min_periods=1).median()

        # Normalize Data
        normalizedDF = SignalAnalysis.normalizeDFofSignalGroup(myDFfiltered, [signalNames])
        normalizedDFRef = SignalAnalysis.normalizeDFofSignalGroup(myDFfilteredRef, [signalNames])

        rowFeatureC2CDf = pd.DataFrame(index = [rowFeatureDf.index])
        for signalName in signalNames:
            # Double the normal Time Constant
            if not useGlobalEnvelopeDecayTimeConstant:
                useGlobalEnvelopeDecayTimeConstant = 2*rowFeatureDf[signalName + '_energyApproach'].values[0]

            normalizationCoefficient = 1 / (myDFfiltered[signalName].max() - myDFfiltered[signalName].min())
            rowFeatureC2CDf[
                signalName + '_c2c_doubleTCEnvelope_globallynormalized'] = SignalAnalysis.calculateCurve2CurveComparison(
                normalizedDF[signalName], normalizedDFRef[signalName], envelopeAmplitude * normalizationCoefficient,
                useGlobalEnvelopeDecayTimeConstant)

        return rowFeatureC2CDf

    @staticmethod
    def preProcessDischarge(df):
        dfSync = SignalUtilities.synchronizeDataframe(df)
        dfSyncDecay = ExponentialSignalAnalysis.getDfwithDecayOnly(dfSync)
        # # Filter Data
        dfSyncDecayFiltered = dfSyncDecay.rolling(window=3, min_periods=1).median().dropna(how='all')
        dfSyncDecayFilteredResampled = SignalAnalysis.resampleSeries(dfSyncDecayFiltered)
        return SignalUtilities.convertIndicesToSeconds(dfSyncDecayFilteredResampled)

    @staticmethod
    def extractFeatures(df, eventTime, uNames, iNames, rNames = {}):

        dfFeature = pd.DataFrame(index=[eventTime])

        # Characteristic time of a pseudo-exponential decay
        dfCTime = QuenchHeaterAnalysis.calculateCharacteristicTimeConstant(df, eventTime, iNames + uNames)
        # Initial final value
        dfInitialFinal = QuenchHeaterAnalysis.calculateInitialFinalValue(df, eventTime, iNames + uNames)
        # Initial resistance
        if rNames is not None:
            dfInitialResistance = QuenchHeaterAnalysis.calculateInitialResistance(dfInitialFinal, eventTime,
                                                                              uNames, iNames, rNames)
        else:
            dfInitialResistance = pd.DataFrame(index=[eventTime])
        # Similarity matrix
        # # Normalize Data
        dfNormalized = SignalAnalysis.normalizeDFofSignalGroup(
            deepcopy(df), [uNames, iNames, rNames])
        dfSimilarityMatrix = QuenchHeaterAnalysis.calculateSimilarityMatrix(dfNormalized,
                                                                            timestamp=eventTime,
                                                                            signalNames=[uNames, iNames, rNames])
        return pd.concat([dfFeature, dfCTime, dfInitialFinal, dfInitialResistance, dfSimilarityMatrix], axis=1, sort=False)

    @staticmethod
    def performCurveToCurveComparison(df, dfRef, featuresRef, iNames, rNames, uNames):
        rowFeatureC2CDf = pd.DataFrame(index=[featuresRef.index])
        c2cVoltage = QuenchHeaterAnalysis.performCurveToCurveComparisonGroup(deepcopy(df[uNames]),
                                                                             deepcopy(dfRef[uNames]),
                                                                             featuresRef, uNames, 20)

        c2cCurrent = QuenchHeaterAnalysis.performCurveToCurveComparisonGroup(
            deepcopy(df[iNames]),
            deepcopy(dfRef[iNames]),
            featuresRef, iNames, 5)

        c2cResistance = QuenchHeaterAnalysis.performCurveToCurveComparisonGroup(
            deepcopy(df[rNames]),
            deepcopy(dfRef[rNames]),
            featuresRef, rNames, 0.5, useGlobalEnvelopeDecayTimeConstant=0.08)

        return pd.concat([rowFeatureC2CDf, c2cVoltage, c2cCurrent, c2cResistance], axis=1)

