from scipy.optimize import leastsq
import numpy as np
import pandas as pd
from .SignalAnalysis import SignalAnalysis
from scipy import signal as s


class ExponentialSignalAnalysis:

    def calculateCharacteristicTimeOfExponentialDecayInSecondsWithIntegral(exponentialDecayInSeconds, exponent):
        return exponent * np.trapz(exponentialDecayInSeconds, x=exponentialDecayInSeconds.index, axis=0) / (
                    exponentialDecayInSeconds.iloc[0] - exponentialDecayInSeconds.iloc[-1])

    def chargeApproach(sliceDF):
        exponent = 1
        sliceSquaredDF = sliceDF.pow(exponent)
        return ExponentialSignalAnalysis.calculateCharacteristicTimeOfExponentialDecayInSecondsWithIntegral(sliceSquaredDF, exponent)


    def energyApproach(sliceDF):
        exponent = 2
        sliceSquaredDF = sliceDF.pow(exponent)
        return ExponentialSignalAnalysis.calculateCharacteristicTimeOfExponentialDecayInSecondsWithIntegral(sliceSquaredDF, exponent)

    def linLog(myDF):
        right_exp = lambda p, x: -p[1] * x + p[0]
        err = lambda p, x, y: (right_exp(p, x) - y)

        time = myDF.index
        rate = np.log(myDF.values)
        v0 = [0., time.min()]
        out = leastsq(err, v0, args=(time, rate))
        v = out[0]  # fit parameters out []
        timeConstant = 1 / v[1]
        return timeConstant

    # put outside of the class?
    def expFit(myDF):
        right_exp = lambda p, x: p[0] * np.exp(-p[2] * (x - p[1]))
        err = lambda p, x, y: (right_exp(p, x) - y)

        time = myDF.index
        rate = myDF.values
        v0 = [0., 0., 0.]
        out = leastsq(err, v0, args=(time, rate))
        v = out[0]  # fit parameters out []
        timeConstant = 1 / v[2]
        return timeConstant

    __nameToFunction = {"chargeApproach": chargeApproach,
                        "energyApproach": energyApproach,
                        "linReg": linLog,
                        "expFit": expFit}

    @staticmethod
    def getCharacteristicTimeOfExponentialDecay(df, functionNamesToApply):
        """

            **returns dataframe with the time constants, calculated with the chosen methods**

            Function returns a dataframe with time constant according to the input method chosen for the passed DataFrame object
            Note that passed DataFrame object should just contain the decay, the functions getWindowWithDecayOnly or
            getWindowWithGivenMeanAndStd can be used for filtering

            :type df: DF
            :type functionNamesToApply: list
            :return: DF

            - Example :

            from source.signals.SignalAnalysis import SignalAnalysis
            import pandas as pd
            import numpy as np

            x = np.arange(100)*0.001
            timeConstant = 0.8
            df = pd.DataFrame(data = np.exp(-x/timeConstant), index = x, columns = ["Voltage_1"])

            functionNamesToApply = ["chargeApproach", "energyApproach", "linReg", "eFit"]
            SignalAnalysis.getDecayTimeConstant(df, functionNamesToApply)

            >>> 					Voltage_1
            >>>chargeApproach		0.8
            >>>energyApproach		0.8
            >>>linReg				0.8
            >>>eFit					0.8

        """

        functionsToApply = [ExponentialSignalAnalysis.__nameToFunction[x] for x in functionNamesToApply]
        if isinstance(df, pd.DataFrame) and not df.empty:
            return df.aggregate(functionsToApply)
        else:
            return df

    @staticmethod
    def calculateChangeInCharacteristicOfExponentialDecay(originSeries, window=100):
        """Method returns the change in the characteristic of the exponential decay over time as a series"""
        # resampledSeries=SignalAnalysis.resampleSeries(originSeries)

        samplingTime = (originSeries.index[1] - originSeries.index[0])
        difSeries = originSeries.diff(periods=window) / samplingTime / window
        tauSeries = -originSeries.div(difSeries)
        return tauSeries

    @staticmethod
    def findStartIndexOfExponentialDecay(expDf):
        # find start index based on current
        decayStartIndex = ExponentialSignalAnalysis.getWindowWithGivenMeanAndStd(expDf['I_HDS_1'])
        if decayStartIndex != 0:
            return (decayStartIndex, 'current')

        # find start index based on voltage correlation to an ideal decay

        # find start index based on voltage decrease
        decayStartIndex = ExponentialSignalAnalysis.getWindowWithDeviationFromInitialValue(expDf['U_HDS_1'])
        if decayStartIndex != 0:
            return (decayStartIndex, 'voltage')

        # if not found, then take the full dataframe
        return (0, 'failed')

    @staticmethod
    def getDfwithDecayOnly(expDf):
        """" Method looks for beginning of the decay and returns DF with only the decay """
        decayStartIndex, source = ExponentialSignalAnalysis.findStartIndexOfExponentialDecay(expDf)

        return expDf.iloc[(decayStartIndex + 1):len(expDf)]

    @staticmethod
    def getWindowWithGivenMeanAndStd(originSeries, indexIncrement=20, mean=50, std=0.1):
        """Method returns a series within a given mean and standart deviation"""
        for j in range(int(len(originSeries) / 10) - indexIncrement):
            window = originSeries.iloc[j:j + indexIncrement]
            windowMean = window.mean()
            windowStd = window.std()
            isWindowFound = (windowMean >= mean) and (windowStd <= std)
            if isWindowFound:
                return j
        return 0

    @staticmethod
    def getWindowWithDeviationFromInitialValue(originSeries, initialValue=0, deviationAllowance=0.998):
        """Method applies low pass filter to series"""
        if initialValue == 0:
            initialValue = np.mean(s.medfilt(originSeries.values[0:19], 3))
        # use filter to find start of decay
        filteredSeries = SignalAnalysis.applyFrequencyFilter(originSeries)
        # get Df with initial values
        initialSeries = filteredSeries[filteredSeries.values > initialValue * deviationAllowance]
        if len(initialSeries) > 0:
            return np.flatnonzero(originSeries.index == initialSeries.index[-1])[0]
        else:
            return 0