from ..cals.CalsDbSignal import CalsDbSignal as CalsDbSignal
from ..post_mortem.PmDbSignal import PmDbSignal as PmDbSignal
from ..influx.InfluxDbSignal import InfluxDbSignal as InfluxDbSignal
from .SignalUtilities import SignalUtilities
import pandas as pd
from copy import deepcopy

class Signal(object):
    """Signal class encapsulates behavior for working with signal databases. It provides read and write methods for accessing
    Post Mortem, CALS and InfluxDB APIs."""

    __timeUnits = {'PM': 'ns', 'CALS': 's', 'INFLUX': None}

    def __init__(self, df=None):
        
        self.__dfState = None
        self.__df = pd.DataFrame()
        if df is not None:
            self.df = df

    #region df property

    @property
    def df(self):
        return self.__df

    @df.setter
    def df(self, df):
        """Method validates that the passed DataFrame"""
        if not isinstance(df, pd.DataFrame):
            raise TypeError(
                "You can only assign a pandas DataFrame to signal.df")

        self.__df = df

    #endregion
    def read(self, api, append=False, **kwargs):
        """
            **Converts input time variable into a unix timestamp integer type**

            Method allowing users to read signal from a given database api

            :param api: input signal database name ("test_cals", "pm", "influx")
            :type api: string
            :param append: flag indicating whether the signal should be appended to the object's dataframe (True) or not (False). False emulates static behavior
            :type append: boolean
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: pandas.dataframe

            - Example CALS::

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  arcCurrentName = "RPTE.UA47.RB.A45:I_MEAS"
                  startTime = 1426220459491000000
                  endTime =   1426220479491000000
                  I_MEAS = Signal().read('test_cals', name=arcCurrentName, startTime = startTime, endTime = endTime, ldb = ldb)
                  print(I_MEAS.head())
                  >>>               RPTE.UA47.RB.A45:I_MEAS
                  >>> 1.426220e+09                  10071.5
                  >>> 1.426220e+09                  10076.5
                  >>> 1.426220e+09                  10081.5
                  >>> 1.426220e+09                  10086.5
                  >>> 1.426220e+09                  10091.5

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  arcCurrentName = "RPTE.UA47.RB.A45:I_MEAS"
                  startTime = 1426220459491000000
                  duration = [(20, 's')]
                  I_MEAS = Signal().read('test_cals', name=arcCurrentName, startTime = startTime, duration = duration, ldb = ldb)
                  print(I_MEAS.head())
                  >>>               RPTE.UA47.RB.A45:I_MEAS
                  >>> 1.426220e+09                  10071.5
                  >>> 1.426220e+09                  10076.5
                  >>> 1.426220e+09                  10081.5
                  >>> 1.426220e+09                  10086.5
                  >>> 1.426220e+09                  10091.5

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  arcCurrentName = "RPTE.UA47.RB.A45:I_MEAS"
                  startTime = 1426220469491000000
                  duration = [(10, 's'), (10, 's')]
                  I_MEAS = Signal().read('test_cals', name=arcCurrentName, startTime = startTime, duration = duration, ldb = ldb)
                  print(I_MEAS.head())
                  >>>               RPTE.UA47.RB.A45:I_MEAS
                  >>> 1.426220e+09                  10071.5
                  >>> 1.426220e+09                  10076.5
                  >>> 1.426220e+09                  10081.5
                  >>> 1.426220e+09                  10086.5
                  >>> 1.426220e+09                  10091.5

            - Expected Fail Response CALS::

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  arcCurrentName = "RPTE.UA47.RB.A45:I_MEAS"
                  startTime = 1426220469491000000
                  duration = [(10, 's'), (10, 's')]
                  I_MEAS = Signal().read('test_cals', name=arcCurrentName, startTime = startTime, duration = duration)
                  print(I_MEAS.head())
                  >>> Exception: You have to provide a instance of pytimber.LoggingDB via the ldb keyword

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  arcCurrentName = "RPTE.UA47.RB.A45:I_MEA"
                  startTime = 1426220469491000000
                  duration = [(10, 's'), (10, 's')]
                  I_MEAS = Signal().read('test_cals', name=arcCurrentName, startTime = startTime, duration = duration, ldb = ldb)
                  print(I_MEAS.head())
                  >>> WARNING:pytimber.pytimber:No variables found.
                  >>> Empty DataFrame
                  >>> Columns: []
                  >>> Index: []
                  >>> UserWarning: Querying CALS with the parameters name: RPTE.UA47.RB.A45:I_MEA, start time: 1426220459.4910002 and end time: 1426220479.4910002 returned no data.
                  >>> An empty DataFrame was returned.
                  >>> ___________________________________________________________________________________
                  >>> Reproduce the error using the raw pytimber API:

                  >>> import pytimber
                  >>> ldb = pytimber.LoggingDB()
                  >>> ldb.get(RPTE.UA47.RB.A45:I_MEA, 1426220459.4910002, 1426220479.49100

            - Example PM::

                  from lhcsmapi.dbsignal.Signal import Signal
                  system = "QPS"
                  className = "DQAMCNMB_PMHSU"
                  magnet = "B20L5"
                  signalName = "I_HDS_4"
                  timestamp = 1426220469491000000
                  I_HDS_4 = Signal().read('pm', system=system, className=className, source=magnet,
                                                signalName=signalName, eventTime=timestamp)
                  print(I_HDS_4.head())
                  >>>                       I_HDS_4
                  >>> 1426220469491020832  0.089648
                  >>> 1426220469491026040  0.091555
                  >>> 1426220469491031248  0.091555
                  >>> 1426220469491036456  0.093463
                  >>> 1426220469491041664  0.091555

            - Expected Fail Response PM::

                      from lhcsmapi.dbsignal.Signal import Signal
                      system = "QPS"
                      className = "DQAMCNMB_PMHSU"
                      magnet = "B20L5"
                      signalName = "I_HDS_4"
                      timestamp = 1426220469491000000
                      I_HDS_4 = Signal().read('pm', system=system, className=className, source=magnet,
                                                    signalName=signalName, eventTime=timestamp)
                      print(I_HDS_4.head())
                      >>> Querying Post Mortem failed using the following query:
                      >>> http://pm-api-pro/v2/pmdata/signal?system=QPS&className=DQAMCNMB_PMHS&source=B20L5&timestampInNanos=1426220469491000000&signal=I_HDS_4
        """

        api = api.upper()

        df = Signal._databaseAccessHandler(api, 'read', **kwargs)
        if append:
            if self.__dfState is not None and self.__dfState != api:
            # if the the recent read request is different from the previous conversions are needed
            # convert both DataFrames
                self.df = SignalUtilities.convertIndicesToTimestamps(self.df, Signal.__timeUnits[self.__dfState])
                df = SignalUtilities.convertIndicesToTimestamps(df, Signal.__timeUnits[api])
                self.__dfState = 'INFLUX'
            else:
                self.__dfState = api
            df = pd.concat([df, self.df], axis=1)
        else:
            self.__dfState = api

        self.df = df

        return self.df

    @staticmethod
    def write(api, **kwargs):
        """
            **Writes df into a database**

            Method writes dataframe associated with an object to a selected database.

            :param api: input signal database name ("influx")
            :type api: string
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary

        """
        return Signal._databaseAccessHandler(api, 'write', **kwargs)

    @staticmethod
    def _databaseAccessHandler(api, access_method_identifier, **kwargs):

        dbAccess = Signal.executeDbSignalFactory(api, **kwargs)
        try:
            access_method = getattr(dbAccess, access_method_identifier, None)
            return access_method(**kwargs)
        except AttributeError:
            raise AttributeError('{} Signal has no {} method'.format(api, access_method_identifier))

    @staticmethod
    def executeDbSignalFactory(signalType, **kwargs):
        """
            **Returns an object of one of the derived DbSignal classes according to the passed signalType**

            Method executes a factory producing signal database APIs.

            :param signalType: input signal database name ("test_cals", "pm", "influx")
            :type signalType: string
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: datafrane

            - Example PM::

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("pm", duration = [(0, 's')])
                  >>> <lhcsmapi.post_mortem.PmDbSignal.PmDbSignal at 0x7f4ec387e780>

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("PM", duration = [(0, 's')])
                  >>> <lhcsmapi.post_mortem.PmDbSignal.PmDbSignal at 0x7f4ec387e780>

            - Example CALS::

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  Signal.executeDbSignalFactory("test_cals", ldb=ldb, duration = [(0, 's')])
                  >>> <lhcsmapi.test_cals.CalsDbSignal.CalsDbSignal at 0x7f4ea980c1d0>

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  Signal.executeDbSignalFactory("CALS", ldb=ldb, duration = [(0, 's')])
                  >>> <lhcsmapi.test_cals.CalsDbSignal.CalsDbSignal at 0x7f4ea980c4e0>

            - Example INFLUX::

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("influx", duration = [(0, 's')])
                  >>> <lhcsmapi.influx.InfluxDbSignal.InfluxDbSignal at 0x7f4ea980c9b0>

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("INFLUX", duration = [(0, 's')])
                  >>> <lhcsmapi.influx.InfluxDbSignal.InfluxDbSignal at 0x7f4ea979a128>

            - Expected Fail Response::

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("nxcals", duration = [(0, 's')])
                  >>> TypeError: NXCALS key does not match any know APIs
                """

        signalType_ = signalType.upper()
        if signalType_ == "PM":
            return PmDbSignal.buildSignal(**kwargs)
        elif signalType_ == "CALS":
            return CalsDbSignal.buildSignal(**kwargs)
        elif signalType_ == "INFLUX":
            return InfluxDbSignal.buildSignal(**kwargs)
        else:
            raise TypeError('{} key does not match any known APIs: pm, test_cals, influx'.format(signalType_))

    def getSynchronizedTimeDataframe(self, t0=None, unit='ns'):
        """
            **Applies the synchronizeDataframe method on a copy of the internal pandas DataFrame**

            Method synchronizes the time index for the object's dataframe.
            It subtracts the initial time bias, i.e., the time starts at 0 s.
            In addition, it expresses time in seconds.

            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :return: pandas.df

            - Example PM (same behaviour for others)::

                from lhcsmapi.dbsignal.Signal import Signal
                system = "QPS"
                className = "DQAMCNMB_PMHSU"
                magnet = "B20L5"
                signalName = "I_HDS_4"
                timestamp = 1426220469491000000
                pmSignal = Signal()
                pmSignal.read('pm', system=system, className=className, source=magnet,
                                            signalName=signalName, eventTime=timestamp)
                I_HDS_4_sync = pmSignal.getSynchronizedTimeDataframe()
                print(I_HDS_4_sync.head())
                >>>            I_HDS_4
                >>> 0.000000  0.089648
                >>> 0.000006  0.091555
                >>> 0.000011  0.091555
                >>> 0.000016  0.093463
                >>> 0.000021  0.091555

            - Expected Fail Response::

                from lhcsmapi.dbsignal.Signal import Signal
                system = "QPS"
                className = "DQAMCNMB_PMHSU"
                magnet = "B20L5"
                signalName = "I_HDS_4"
                timestamp = 1426220469491000000
                pmSignal = Signal()
                pmSignal.read('pm', system=system, className=className, source=magnet,
                                            signalName=signalName, eventTime=timestamp)
                I_HDS_4_sync = pmSignal.getSynchronizedTimeDataframe("s")
                print(I_HDS_4_sync.head())
                >>> OverflowError: Python int too large to convert to C long
        """
        output = SignalUtilities.synchronizeDataframe(deepcopy(self.df), t0)
        return output

    def getSynchronizedTime(self):
        """
            **Returns a list with synchronized seconds**

            Method synchronizes the time index for the object's dataframe and returns the index as a list.
            It subtracts the initial time bias, i.e., the time starts at 0 s.
            In addition, it expresses time in seconds.

            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :return: list

            - Example PM (same behaviour for others)::

                from lhcsmapi.dbsignal.Signal import Signal
                system = "QPS"
                className = "DQAMCNMB_PMHSU"
                magnet = "B20L5"
                signalName = "I_HDS_4"
                timestamp = 1426220469491000000
                pmSignal = Signal()
                pmSignal.read('pm', system=system, className=className, source=magnet,
                                            signalName=signalName, eventTime=timestamp)
               pmSignal.getSynchronizedTime()
               >>> [0.0, 5.0067901611328125e-06,
               >>>  1.049041748046875e-05,
               >>>  1.5497207641601562e-05,
               >>>  2.09808349609375e-05
               >>> ...
        """

        if not self.df.empty and self.df is not None:
            df_ = deepcopy(self.df)
            time = df_.index.tolist()
            time = [t/1e9-time[0]/1e9 for t in time]

            return time
