import influxdb as influxdb
import pandas as pd
from datetime import datetime

from ..dbsignal.DbSignal import DbSignal
from ..dbsignal.SignalUtilities import SignalUtilities as su
from ..Time import Time as tc
import tzlocal
import warnings
import numpy as np


class InfluxDbSignal(DbSignal):
    """Class for encapsulating reading and writing methods of the Influx API"""

    # __conversions = {'float64': 's', 'int64': 'ns'}

    def __init__(self, client, df=None, query=None, tags={}, measurement=None, startTime=None, endTime=None, tz=None):

        super().__init__('Influx')
        self.client = client
        self.df = df

        self.query = query
        self.tags = tags
        self.measurement = measurement
        self.startTime = startTime
        self.endTime = endTime
        self.tz = tz

        self.protocol = 'json'

    @staticmethod
    def createDataFrameClient(username='', password='', host='dbod-lhc-sm.cern.ch', port=8081, database='test',ssl=True):
        """"Method creates and returns a DataFrameClient"""
        return influxdb.DataFrameClient(host, port, username, password, database, ssl)

    @staticmethod
    def buildSignal(**kwargs):
        """Method validates kwargs to create and return a InfluxDbSignal object.
        The method looks though kwargs for arguments to create a new DataFrameClient. If kwargs contains the necessary
        arguments, a client is created. If kwargs contain a DataFrameClient object, this is used instead.
        If neither a complete client or parameters to build one is provided, a default client is created.
        The resulting client is used to create a return a InfluxDbSignal object"""
        client = None

        # client keys:
        username = InfluxDbSignal.validateKeyInkwargs('username', str, **kwargs)
        password = InfluxDbSignal.validateKeyInkwargs('password', str, **kwargs)
        host = InfluxDbSignal.validateKeyInkwargs('host', str, **kwargs)
        port = InfluxDbSignal.validateKeyInkwargs('port', int, **kwargs)
        database = InfluxDbSignal.validateKeyInkwargs('database', str, **kwargs)

        if username and password and host and port and database:
            client = InfluxDbSignal.createDataFrameClient(username, password, host, port, database)
        else:
            client = InfluxDbSignal.createDataFrameClient('root', 'root')

        client = InfluxDbSignal.validateKeyInkwargs("client", influxdb.DataFrameClient, **kwargs) or client

        #df
        df = InfluxDbSignal.validateKeyInkwargs('df', pd.DataFrame, **kwargs)

        # influx query keys:
        query = InfluxDbSignal.validateKeyInkwargs('query', str, **kwargs)
        measurement = InfluxDbSignal.validateKeyInkwargs('measurement', str, **kwargs)

        startTime = InfluxDbSignal.validateKeyInkwargs('startTime', int, pd.Timestamp, str, float, datetime, tuple, **kwargs)
        endTime = InfluxDbSignal.validateKeyInkwargs('endTime', int, pd.Timestamp, str, float, datetime, tuple, **kwargs)
        duration = InfluxDbSignal.validateKeyInkwargs('duration', tuple, list, **kwargs)
        tz = InfluxDbSignal.validateKeyInkwargs('tz', str, kwargs) or tzlocal.get_localzone().__str__()

        tags = InfluxDbSignal.validateKeyInkwargs('tags', dict, **kwargs) or {}

        # calculate start and end time based of passed arguments, if both None, default query with will get all timestamps
        startTime, endTime = tc.getQueryPeriodInUnixTime(startTime, endTime, duration, tz)

        # get the default tags
        defaultTags = {}
        defaultTags['system'] = InfluxDbSignal.validateKeyInkwargs('system', str, **kwargs)
        defaultTags['circuit'] = InfluxDbSignal.validateKeyInkwargs('circuit', str, **kwargs)
        defaultTags['magnet'] = InfluxDbSignal.validateKeyInkwargs('magnet', str, **kwargs)

        # enter default tags into tags if not already present
        for k, v in defaultTags.items():
            if v is not None and k not in tags.keys():
                tags[k] = v

        return InfluxDbSignal(client, df, query, tags, measurement, startTime, endTime, tz)

    def write(self, signal=None, **kwargs):
        """Method writes the DataFrame contained in signal to the database specified in the signal instance's client field.
        Returns True if success"""

        # if not self.df.empty:
        #     dtype = str(self.df.index.dtype)
        #
        #     # if dtype in InfluxDbSignal.__conversions:
        #     #     self.df = su.convertIndicesToTimestamps(self.df, InfluxDbSignal.__conversions[dtype])

        # columns = self.df.columns.tolist()

        # for column in self.df.columns:
        #     if self.df[column].hasnans:
        #         self.df[column] = self.df.replace(np.nan,'NaN')
        #         self.df[column] = self.df[column].apply(str)

        self.df = self.df.dropna(axis='columns')
        self.df = self.df.drop_duplicates()

        try:
            return self.client.write_points(dataframe=self.df,
                                        measurement=self.measurement,
                                        tags=self.tags,
                                        protocol=self.protocol,
                                        time_precision='n')
        except:
            raise Exception('writing the dataframe to the database failed')

    def read(self, signal=None, **kwargs):
        """Method queries the database specified in the client and returns the result as pandas DataFrame"""

        queryString = self.query

        if not queryString:
            queryString = self.createInfluxDbQuery(self.measurement)

        queryResult = None

        try:
            queryResult = self.client.query(queryString)
        except:
            raise Exception("Querying InfluxDB with query: {} failed". format(queryString))

        df = queryResult.get(self.measurement)
        if df is None or df.empty:
            warnings.warn(
                "The query: {}, returned no data. Therefore an empty DataFrame is returned instead".format(queryString))
            df = pd.DataFrame()

        # for column in df.columns:
        #     if 'NaN' in df[column]:
        #         df[column] = self.df.replace('NaN', np.nan)
        #         self.df[column] = self.df[column].apply(np.float64)

        if not df.empty:
            newIndex = df.index.tz_convert(self.tz)
            df = df.set_index([newIndex])
            
        return df

    # region create Query

    def createInfluxDbQuery(self, measurement, *fieldKeys):
        """Method returns a query string based on the internal fields"""

        baseQuery = self.__createBaseQueryString(measurement, *fieldKeys)
        timeConstraint = self.__createTimeConstraintSubstring()
        tagConstraint = self.__createTagConsraintSubstring()

        if tagConstraint or timeConstraint:
            baseQuery += ' where'

        if tagConstraint and timeConstraint:
            timeConstraint = ' and' + timeConstraint

        return baseQuery + tagConstraint + timeConstraint

    def __createBaseQueryString(self, measurement, *fieldKeys):
        """Method returns a base query describing what fields to get from a certain measurement"""
        selectedFields = ''


        # set the fields to select
        numOfFieldKeys = len(fieldKeys)
        if numOfFieldKeys > 0:
            for i in range(numOfFieldKeys):
                selectedFields += fieldKeys[i]
                if i != numOfFieldKeys - 1:
                    selectedFields += ', '
        else:
            selectedFields += '*'

        # create base query - default: 'select * from {measuremenent}'
        return 'select ' + selectedFields + ' from ' + measurement

    def __createTimeConstraintSubstring(self):
        """Method creates a substring containing the time constraints for a influxDB query"""
        timeConstraintString = ''

        if self.startTime:
            startTime = tc(self.startTime, 'ns', 'UTC').stringTimestamp[:-7]
            timeConstraintString += " time >= \'{}\'".format(startTime)

        if self.startTime and self.endTime:
            endTime = tc(self.endTime, 'ns', 'UTC').stringTimestamp[:-7]
            timeConstraintString += " and time <= \'{}\'".format(endTime)
            
        return timeConstraintString

    def __createTagConsraintSubstring(self):
        """Method creates a substring containing the tag constraints for a influxDB query"""

        tagsConstraintString = ''
        tagSet = [(k, v) for k, v in self.tags.items()]

        numberOfTags = len(tagSet)
        for i in range(numberOfTags):
            k, v = tagSet[i]
            tagsConstraintString += ' {} = \'{}\''.format(k, v)
            if i < numberOfTags - 1:
                tagsConstraintString += ' and'

        return tagsConstraintString

    #endregion
