import os
import pandas as pd
import numpy as np
from functools import reduce

class MappingMetadata(object):

    @staticmethod
    def getCrateToMagnetAndVoltageFeelerMapping() -> pd.DataFrame:
        # https://wikis.cern.ch/display/MPEEP/DQLPUS+crates+and+magnets
        # only magnets which have TD in the table entries (trigger lines)
        selectedPath = os.path.dirname(__file__)
        fullPath = os.path.join(selectedPath, 'qps_crate/RB_CrateToMagnetAndVoltageFeelerMap.csv')
        return pd.read_csv(fullPath)

    @staticmethod
    def getMagnetToCrateAndVoltageFeelerMapping(circuit_type) -> pd.DataFrame:
        # https://wikis.cern.ch/display/MPEEP/DQLPUS+crates+and+magnets
        # only magnets which have TD in the table entries (trigger lines)
        selectedPath = os.path.dirname(__file__)
        if circuit_type == 'RB':
            fullPath = os.path.join(selectedPath, 'qps_crate/RB_MagnetToCrateAndVoltageFeelerMap.csv')
        elif circuit_type == 'RQD':
            fullPath = os.path.join(selectedPath, 'qps_crate/RQD_MagnetToCrateMap.csv')
        elif circuit_type == 'RQF':
            fullPath = os.path.join(selectedPath, 'qps_crate/RQF_MagnetToCrateMap.csv')
        return pd.read_csv(fullPath)

    @staticmethod
    def getMagnetCellCrateCircuitNameMapping() -> pd.DataFrame:
        selectedPath = os.path.dirname(__file__)
        fullPath = os.path.join(selectedPath, 'magnet/RB_MagnetCellQpscrateSector.csv')
        return pd.read_csv(fullPath)

    @staticmethod
    def getBusbarToMagnetMapping(circuitType):
        if circuitType == "RB":
            fileName = "busbar/nQPS_RB_busBarProtectionDetails.csv"
        elif any(rqType == circuitType for rqType in ["RQD", "RQF"]):
            fileName = "busbar/nQPS_RQ_busBarProtectionDetails.csv"
        else:
            raise KeyError("Circuit type {} is not supported".format(circuitType))
        selectedPath = os.path.dirname(__file__)
        fullPath = os.path.join(selectedPath, fileName)
        return pd.read_csv(fullPath)

    @staticmethod
    def getMagnetsForCircuitNames(circuitType, circuitNames):
        circuitNames = [circuitNames] if type(circuitNames) is not list else circuitNames  # to a decorator!

        if circuitType == "RB":
            magnetToCircuitName = MappingMetadata.getMagnetCellCrateCircuitNameMapping()
            areMagnetsMatchingCircuitName = magnetToCircuitName["Sector"].isin(circuitNames)
            magnetsMatchingCircuitName = magnetToCircuitName[areMagnetsMatchingCircuitName]
            return list(magnetsMatchingCircuitName['Magnet'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuitType))

    @staticmethod
    def getCellsForCircuitNames(circuitType, circuitNames):
        circuitNames = [circuitNames] if type(circuitNames) is not list else circuitNames  # to a decorator!

        if circuitType == "RB":
            magnetToCircuitName = MappingMetadata.getMagnetCellCrateCircuitNameMapping()
            areMagnetsMatchingCircuitName = magnetToCircuitName["Sector"].isin(circuitNames)
            magnetsMatchingCircuitName = magnetToCircuitName[areMagnetsMatchingCircuitName]
            return list(magnetsMatchingCircuitName['Cell'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuitType))

    # To be removed
    @staticmethod
    def getCrateNameFromMagnetName(circuit_type: str, magnet: str) -> str:
        magnets_crates = MappingMetadata.getMagnetToCrateAndVoltageFeelerMapping(circuit_type)
        magnets_crates.fillna(value='', inplace=True)
        if circuit_type == 'RB':
            mask = reduce(np.logical_or,
                                    [magnets_crates['Magnet{}'.format(i)].str.contains(magnet) for i in range(1, 5)])
        elif any([ct == circuit_type for ct in ['RQD', 'RQF']]):
            mask = magnets_crates.apply(lambda row: magnet in row['U_DIODE_{}'.format(circuit_type)], axis=1)

        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))
        return magnets_crates.loc[mask]['Crate'].values[0]

    @staticmethod
    def getCratesForCircuitNames(circuitType, circuitNames):
        circuitNames = [circuitNames] if type(circuitNames) is not list else circuitNames  # to a decorator!

        if circuitType == "RB":
            crateToMagnet = MappingMetadata.getCrateToMagnetAndVoltageFeelerMapping()
            isCrateMatchingSector = crateToMagnet["Sector"].isin(circuitNames)
            cratesMatchingSector = crateToMagnet[isCrateMatchingSector]
            return list(cratesMatchingSector['Crate'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuitType))

    @staticmethod
    def getVoltageFeelersForCircuitNames(circuitType, circuitNames):
        circuitNames = [circuitNames] if type(circuitNames) is not list else circuitNames  # to a decorator!

        if circuitType == "RB":
            crateToMagnet = MappingMetadata.getCrateToMagnetAndVoltageFeelerMapping()
            isCrateMatchingSector = crateToMagnet["Sector"].isin(circuitNames)
            cratesMatchingSector = crateToMagnet[isCrateMatchingSector]
            return list(cratesMatchingSector['U_EARTH'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuitType))

    @staticmethod
    def getBusbarsForCircuitName(circuitType, circuitNames):
        circuitNames = [circuitNames] if type(circuitNames) is not list else circuitNames # to a decorator!

        bbDf = MappingMetadata.getBusbarToMagnetMapping(circuitType)
        isBusbarMatchingCircuitName = bbDf['Sector'].isin(circuitNames)
        busbarMatchingCircuitName = bbDf[isBusbarMatchingCircuitName]['Bus Bar Segment Name'].values
        return list(busbarMatchingCircuitName)

    @staticmethod
    def getCrateNameFromBusbarNames(circuitType, busbarNames):
        busbarNames= [busbarNames] if type(busbarNames) is not list else busbarNames

        bbDf = MappingMetadata.getBusbarToMagnetMapping(circuitType)
        isBusbarMatchingCircuitName = bbDf['Bus Bar Segment Name'].isin(busbarNames)
        busbarMatchingCircuitName = bbDf[isBusbarMatchingCircuitName]['QPS Crate'].values
        return list(busbarMatchingCircuitName)
