import json
import os
import re
import pandas as pd

from lhcsmapi.metadata.MappingMetadata import MappingMetadata


class SignalMetadata(object):
    BMODE = "HX:BMODE"

    __paths = {'RB': 'RB_METADATA.json',
               'RQF': 'RQF_METADATA.json',
               'RQD': 'RQD_METADATA.json',
               'IPD2_B1B2': 'IPD2_B1B2_METADATA.json',
               'IPD2': 'IPD2_METADATA.json',
               'IPQ2': 'IPQ2_METADATA.json',
               'IPQ4': 'IPQ4_METADATA.json',
               'IPQ8': 'IPQ4_METADATA.json',
               '600A': '600A_METADATA.json'}

    @staticmethod
    def parseMetadata(metadataKey):
        selectedPath = os.path.dirname(__file__)
        try:
            selectedPath = os.path.join(selectedPath, SignalMetadata.__paths[metadataKey.upper()])
        except KeyError:
            raise KeyError("{} is not mapped to a json file".format(metadataKey))

        return SignalMetadata.__parseJsonFile(selectedPath)

    @staticmethod
    def getBeamModeDetails():
        # https://lhc-commissioning.web.cern.ch/lhc-commissioning/systems/data-exchange/published-topics/beam-modes.htm
        selectedPath = os.path.dirname(__file__)
        fullPath = os.path.join(selectedPath, 'beam_mode/beam_mode.csv')
        return pd.read_csv(fullPath, sep=';')

    @staticmethod
    def __parseJsonFile(fileName):
        try:
            with open(fileName) as json_file:
                return json.load(json_file)
        except:
            raise FileNotFoundError("Could not load {}".format(fileName))

    @staticmethod
    def getCircuitTypes():
        return list(SignalMetadata.__paths.keys())

    @staticmethod
    def getCircuitNames(circuitType):
        circuitMetadata = SignalMetadata.parseMetadata(circuitType)
        return list(circuitMetadata["CIRCUITS"].keys())

    @staticmethod
    def getSystemTypesPerCircuit(circuitType):
        circuitMetadata = SignalMetadata.parseMetadata(circuitType)
        circuitName = list(circuitMetadata["CIRCUITS"].keys())[0]
        types = []
        for key, value in circuitMetadata["CIRCUITS"][circuitName].items():
            if type(value) is dict:
                types.append(value['TYPE'])
            elif type(value) is list:
                for element in value:
                    types.append(element['TYPE'])
        return types

    @staticmethod
    def getSectorNamesFromComponents(identifier, system, *keys):
        if len(keys) == 1 and isinstance(keys[0], list):
            keys = keys[0]

        try:
            metadata = SignalMetadata.parseMetadata(identifier)['REL_IP_POS_TO_CIRCUIT_NAME']
            if any([system == sys for sys in ["MAGNET", "BUSBAR"]]):
                regExp = SignalMetadata.parseMetadata(identifier)[system]['REL_IP_POS_REG_EXP']
            elif system == "PC":
                regExp = SignalMetadata.parseMetadata(identifier)[system]['REL_IP_POS_REG_EXP']
                if len(keys) == 1:
                    return re.findall(regExp, keys[0])[0]
                else:
                    return [re.findall(regExp, key)[0] for key in keys]

            return SignalMetadata.__getArcNamesFromMagnetNames(metadata, regExp, *keys)
        except KeyError as e:
            raise KeyError(str(e) + "in object parsed from {}".format(SignalMetadata.__paths[identifier]))

    @staticmethod
    def __getArcNamesFromMagnetNames(metadata, regExp, *keys):
        if len(keys) == 1:
            return SignalMetadata.__getSectorNameFromMagnetName(metadata, regExp, keys[0])
        else:
            return [SignalMetadata.__getSectorNameFromMagnetName(metadata, regExp, key) for key in keys]

    @staticmethod
    def __getSectorNameFromMagnetName(metadata, regExp, key):
        foundRegExp = re.findall(regExp, key)
        if len(foundRegExp) != 1:
            raise ValueError("Key {} does not contain proper MB magnet name".format(key))
        else:
            key = foundRegExp[0]
            if key not in metadata.keys():
                raise KeyError("Key {} does not exist ".format(key))
            else:
                return metadata[key]

    @staticmethod
    def getCircuitNameMetadata(circuitType, circuitName):
        circuitMetadata = SignalMetadata.parseMetadata(circuitType)
        circuitNames = circuitMetadata["CIRCUITS"].keys()
        if circuitName in circuitNames:
            return circuitMetadata["CIRCUITS"][circuitName]
        else:
            raise KeyError("Circuit name {} not present in {}.".format(circuitName, circuitNames))

    @staticmethod
    def getCircuitVariableDatabaseMetadata(circuitType, circuitName, systemType, databaseName):
        circuitMetadata = SignalMetadata.parseMetadata(circuitType)

        # Check if variable type is available
        systemTypes = SignalMetadata.getSystemTypesPerCircuit(circuitType)
        if systemType in systemTypes:
            circuitVariableTypeMetadata = circuitMetadata[systemType]
        else:
            raise KeyError("Variable type {} not present in {}.".format(systemType, systemTypes))

        # Check if databaseName is available
        databaseNames = circuitVariableTypeMetadata.keys()
        if databaseName in databaseNames:
            # update the source wildcard
            if databaseName == "PM":
                source = circuitVariableTypeMetadata[databaseName]['source']
                circuitNameMetadata = SignalMetadata.getCircuitNameMetadata(circuitType, circuitName)
                if source in circuitNameMetadata.keys():
                    circuitVariableTypeMetadata[databaseName]['source'] = circuitNameMetadata[source]
                elif source in circuitNameMetadata[systemType].keys():
                    circuitVariableTypeMetadata[databaseName]['source'] = circuitNameMetadata[systemType][source]

            return circuitVariableTypeMetadata[databaseName]
        else:
            raise KeyError("Database name {} not present in {}.".format(databaseName, databaseNames))

    @staticmethod
    def getVariableNames(circuitType, circuitName, systemType, databaseName):
        circuitVariableTypeDbMetadata = SignalMetadata.getCircuitVariableDatabaseMetadata(circuitType, circuitName,
                                                                                          systemType, databaseName)
        if databaseName == "PM":
            outputList = list(circuitVariableTypeDbMetadata.keys() - {"system", "className", "source"})
        else:
            outputList = list(circuitVariableTypeDbMetadata.keys())

        outputList.sort()
        return outputList

    @staticmethod
    def getVariableName(circuitType, circuitName, systemType, databaseName, variableName):
        if type(circuitName) is not list:
            circuitName = [circuitName]

        variableName = variableName if type(variableName) is list else [variableName]

        variableNames = []
        for cn in circuitName:
            for vn in variableName:
                varNames = SignalMetadata.updateWildCardsInVariableNamePerCircuitName(cn, circuitType, databaseName,
                                                                                      systemType, vn)
                if type(varNames) is list:
                    variableNames += varNames
                else:
                    variableNames.append(varNames)

        if len(variableNames) == 1:
            return variableNames[0]
        else:
            return variableNames

    @staticmethod
    def updateWildCardsInVariableNamePerCircuitName(circuitName, circuitType, databaseName, systemType, variableName):
        circuitVariableTypeDbMetadata = SignalMetadata.getCircuitVariableDatabaseMetadata(circuitType, circuitName,
                                                                                          systemType, databaseName)
        variableName = circuitVariableTypeDbMetadata[variableName]
        # if variableName is a list
        if type(variableName) is list:
            return [SignalMetadata.updateWildCardsInVariableName(circuitName, circuitType, systemType, vn) for vn in
                    variableName]
        # if variableName is a scalar
        else:
            return SignalMetadata.updateWildCardsInVariableName(circuitName, circuitType, systemType, variableName)

    @staticmethod
    def updateWildCardsInVariableName(circuitName, circuitType, systemType, variableName):
        # get all wildcars to be replaced
        circuitNameMetadata = SignalMetadata.getCircuitNameMetadata(circuitType, circuitName)
        # # for circuitName
        for key, value in circuitNameMetadata.items():
            if "%" in key:
                variableName = variableName.replace(key, value)
        # # for systemType
        systemTypeMetadata = circuitNameMetadata[systemType]
        for key, value in systemTypeMetadata.items():
            if key != "TYPE":
                if type(value) is list:
                    return [variableName.replace(key, val) for val in value]
                else:
                    return variableName.replace(key, value)
        return variableName

    @staticmethod
    def getDatabaseNames(circuitType, systemType):
        circuitMetadata = SignalMetadata.parseMetadata(circuitType)

        systemTypes = circuitMetadata.keys()
        if systemType in systemTypes:
            systemTypeMetadata = circuitMetadata[systemType]
        else:
            raise KeyError("System type {} not present in {}.".format(systemType, systemTypes))

        databaseNames = []
        for key, value in systemTypeMetadata.items():
            if any(db in key for db in ["PM", "CALS"]) and (len(value) > 0):
                databaseNames.append(key)

        return databaseNames

    @staticmethod
    def getMagnetsForCircuitNames(circuitType, circuitNames):
        if type(circuitNames) is not list:
            circuitNames = [circuitNames]

        if circuitType == "RB":
            magnetToCircuitName = MappingMetadata.getMagnetCellCrateCircuitNameMapping()
            areMagnetsMatchingCircuitName = magnetToCircuitName["Sector"].isin(circuitNames)
            magnetsMatchingCircuitName = magnetToCircuitName[areMagnetsMatchingCircuitName]
            return list(magnetsMatchingCircuitName['Magnet'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuitType))

    @staticmethod
    def getCellsForCircuitNames(circuitType, circuitNames):
        if type(circuitNames) is not list:
            circuitNames = [circuitNames]

        if circuitType == "RB":
            magnetToCircuitName = MappingMetadata.getMagnetCellCrateCircuitNameMapping()
            areMagnetsMatchingCircuitName = magnetToCircuitName["Sector"].isin(circuitNames)
            magnetsMatchingCircuitName = magnetToCircuitName[areMagnetsMatchingCircuitName]
            return list(magnetsMatchingCircuitName['Cell'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuitType))

    @staticmethod
    def getCircuitNameFromMagnetOrCrate(circuitType, magnetOrCrate):
        if circuitType == 'RB':
            regExp = SignalMetadata.parseMetadata(circuitType)['MAGNET']['REL_IP_POS_REG_EXP']
            suffixToSectorName = SignalMetadata.parseMetadata(circuitType)['REL_IP_POS_TO_CIRCUIT_NAME']
            foundRegExp = re.findall(regExp, magnetOrCrate)
            return suffixToSectorName[foundRegExp[0]]
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuitType))

    @staticmethod
    def getWildcardForCircuitSystemDbSignal(circuit, circuitName, system, db, signal):
        if any([system == sys for sys in ['QH', 'QDS']]) and circuit == 'RB':
            if db == "PM":
                wildcard = MappingMetadata.getCellsForCircuitNames(circuit, circuitName)
                return ('Cell', wildcard)
            elif db == 'CALS':
                wildcard = MappingMetadata.getMagnetsForCircuitNames(circuit, circuitName)
                return ('Magnet', wildcard)
            else:
                raise KeyError("Database type {} not present in {}.".format(db, ['PM', 'CALS']))
        elif system == 'VF' and circuit == 'RB':
            wildcard = MappingMetadata.getVoltageFeelersForCircuitNames(circuit, circuitName)
            return ('Voltage feeler', wildcard)
        elif system == "BUSBAR":
            wildcard = MappingMetadata.getBusbarsForCircuitName(circuit, circuitName)
            return ('Busbar', wildcard)
        elif system == 'DIODE_RB' and circuit == 'RB':
            if signal == 'U_REF_N1':
                wildcard = MappingMetadata.getCratesForCircuitNames(circuit, circuitName)
                return ('Crate', wildcard)
            else:
                wildcard = MappingMetadata.getMagnetsForCircuitNames(circuit, circuitName)
                return ('Magnet', wildcard)
        return []

    @staticmethod
    def updatePMSourceWithWildcard(sourceIn, circuit, system, signal, wildcard):
        if type(wildcard) is not list:
            wildcard = [wildcard]

        if any([system == sys for sys in ['QH', 'QDS']]):
            return wildcard
        elif system == 'BUSBAR':
            return MappingMetadata.getCrateNameFromBusbarNames(circuit, wildcard)
        elif 'DIODE' in system:
            if signal == 'U_REF_N1':
                return wildcard
            else:
                return [MappingMetadata.getCrateNameFromMagnetName(circuit, wc) for wc in wildcard]
        elif system == 'VF':
            return [MappingMetadata.getCrateNameFromMagnetName(circuit, wc) for wc in wildcard]
        # for VF, PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return sourceIn

    @staticmethod
    def updatePMSignalNamesWithWildcard(signalNamesIn, system, signal, wildcard):
        if type(wildcard) is not list:
            wildcard = [wildcard]
        if type(signalNamesIn) is not list:
            signalNamesIn = [signalNamesIn]

        if system == 'QDS':
            return [signalName.replace("%CELL%", wc) for wc in wildcard for signalName in signalNamesIn]
        elif system == 'BUSBAR':
            return [signalName.replace("%BUSBAR%", wc) for wc in wildcard for signalName in signalNamesIn]
        elif 'DIODE'  in system:
            if signal == 'U_REF_N1':
                return [signalName.replace("%QPS_CRATE%", wc) for wc in wildcard for signalName in signalNamesIn]
            else:
                return [signalName.replace("%MAGNET%", wc) for wc in wildcard for signalName in signalNamesIn]
        elif system == 'VF':
            return [signalName.replace("%MAGNET%", wc) for wc in wildcard for signalName in signalNamesIn]
        # for PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return signalNamesIn

    @staticmethod
    def updateCALSSignalNamesWithWildcard(signalNamesIn, system, wildcard):
        if type(wildcard) is not list:
            wildcard = [wildcard]
        if type(signalNamesIn) is not list:
            signalNamesIn = [signalNamesIn]

        if any([system == sys for sys in ['QH', 'QDS', 'DIODE', 'VF']]):
            return [signalName.replace("%MAGNET%", wc) for wc in wildcard for signalName in signalNamesIn]
        elif system == 'BUSBAR':
            return [signalName.replace("%BUSBAR%", wc) for wc in wildcard for signalName in signalNamesIn]
        # for PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return signalNamesIn