import pandas as pd
import numpy as np
import warnings

from ..Time import Time
from ..post_mortem.PmDbRequest import PmDbRequest
from ..dbsignal.DbSignal import DbSignal


class PmDbSignal(DbSignal):
    """PmDbSignal class encapsulates parsing of a signal obtained from the Post Mortem."""

    def __init__(self, system={}, className={}, source={}, signalName={}, eventTime={}, startTime={}, intervTime={},
                 unit={}, time={}, value={}, pmRestApiPath={}):
        super().__init__('PM')
        self.System = system
        self.ClassName = className
        self.Source = source
        self.SignalName = signalName
        self.EventTime = eventTime
        self.StartTime = startTime
        self.IntervTime = intervTime
        self.Unit = unit
        self.Value = value
        self.Time = time
        self.request = None
        self.pmRestApiPath = pmRestApiPath

    def __str__(self):
        return "Db: PM, Signal: {}, eventTime = {} ({}),\
 system={}, source={}, className={}".format(self.SignalName, self.EventTime, Time.convertToString(self.EventTime, 'ns'), self.System, self.Source, self.ClassName)


    @staticmethod
    def buildSignal(**kwargs):
        """
            **Builds a PmDbSignal object**

            Method builds a PmDbSignal from an input (keyword: argument) dictionary.
            It validates input kwargs and implements static builder design pattern to build a new PmDbSignal object.
            If arguments are not provided, default arguments are used

            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: initialized PmDbSignal object

            - Example::

                from lhcsmapi.post_mortem.PmDbSignal import PmDbSignal
                system = "QPS"
                className = "DQAMCNMB_PMHSU"
                source = "B20L5"
                signalName = "I_HDS_4"
                eventTime = 1426220469491000000
                PmDbSignal().buildSignal(system=system, className=className, source=source,
                              signalName=signalName, eventTime=eventTime)
               >>> <lhcsm.post_mortem.PmDbSignal.PmDbSignal at 0x7f7d89ede550>
        """

        system = PmDbSignal.validateKeyInkwargs('system', str, **kwargs) or {}
        className = PmDbSignal.validateKeyInkwargs('className', str, **kwargs) or {}
        source = PmDbSignal.validateKeyInkwargs('source', str, **kwargs) or {}
        signalName = PmDbSignal.validateKeyInkwargs('signalName', str, **kwargs) or {}
        eventTime = PmDbSignal.validateKeyInkwargs('eventTime', int, pd.Timestamp, np.int64, **kwargs) or {}
        intervTime = PmDbSignal.validateKeyInkwargs('intervTime', int, pd.Timestamp, **kwargs) or {}
        unit = PmDbSignal.validateKeyInkwargs('unit', str, **kwargs) or {}
        time = PmDbSignal.validateKeyInkwargs('time', list, **kwargs) or {}
        value = PmDbSignal.validateKeyInkwargs('value', list, **kwargs) or {}
        pmRestApiPath = PmDbSignal.validateKeyInkwargs('pmRestApiPath', str, **kwargs) or "http://pm-api-pro/v2/"

        return PmDbSignal(system, className, source, signalName, eventTime, {}, intervTime, unit, time, value,
                          pmRestApiPath)

    def read(self, signal=None, **kwargs):
        """
            **Reads a signal from PM database**

            Method returns a pandas Dataframe containing a signal with an index formatted to ns.
            It requires the PmDbSignal object to be built with PmDbSignal.buildSignal() method.

            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query (optional)
            :type kwargs: dictionary
            :return: initialized PmDbSignal object

            - Example::

                from lhcsmapi.post_mortem.PmDbSignal import PmDbSignal
                system = "QPS"
                className = "DQAMCNMB_PMHSU"
                source = "B20L5"
                signalName = "I_HDS_4"
                eventTime = 1426220469491000000
                pms = PmDbSignal.buildSignal(system=system, className=className, source=source,
                                              signalName=signalName, eventTime=eventTime)
                I_HDS_4 = pms.read()
                print(I_HDS_4.head())
                >>>                       I_HDS_4
                >>> 1426220469491020832  0.089648
                >>> 1426220469491026040  0.091555
                >>> 1426220469491031248  0.091555
                >>> 1426220469491036456  0.093463
                >>> 1426220469491041664  0.091555
        """

        self._parseSignal()
        df = self._getSignalDataframe(self.SignalName)

        if df.empty:
            warning = """Querying Post Mortem failed using the following query:
            {} returned no data so an empty DataFrame is being returned""".format(self.request)
            warnings.warn(warning)
            df = pd.DataFrame()

        return df

    def _parseSignal(self):
        """Method looks for structure parameter and time in order to return a Dataframe with the signal."""
        response = PmDbRequest.getResponse("pmdata/signal", False, False, self.pmRestApiPath, system=self.System,
                                           className=self.ClassName,
                                           source=self.Source, timestampInNanos=self.EventTime, signal=self.SignalName)

        if "v2" in self.pmRestApiPath:
            content = response['content'][0]
            # find names
            nameOfValue = PmDbSignal._getNamesOfValue(self, content)
            # find time and value
            PmDbSignal._getTimeValue(self, nameOfValue, content)
        elif "v3" in self.pmRestApiPath:
            self.EventTime = response['content'][0]['header']['timestamp']
            self.Time = response['content'][0]['values']['circ.AQNTIME']['value']
            self.Value = response['content'][0]['values'][self.SignalName]['value']
        return self

    def _getSignalDataframe(self, *columns):
        """Method returns a pandas DataFrame based on internal time and value properties."""
        return pd.DataFrame(self.Value, self.Time, columns=[col for col in columns])

    def _getNamesOfValue(self, content):
        """Method returns a list of names contained in content."""

        size = len(content['namesAndValues'])
        nameOfValue = ["" for x in range(0, size)]
        nameOfValue[:] = [content['namesAndValues'][n]['name'] for n in range(0, size)]

        return nameOfValue

    def _getTimeValue(self, nameOfValue, content):
        """Method sets time (in ns) and value from the content received with a PM signal query."""

        self.EventTime = content['rawDataEntry']['pmEventStamp']
        for s in range(0, len(nameOfValue)):
            if "START_TIME_NSEC" in nameOfValue[s]:  # search for StartTime
                self.StartTime = content['namesAndValues'][s]['value']
            elif "INTERVAL_NSEC" in nameOfValue[s]:  # search for Inteval
                self.IntervTime = content['namesAndValues'][s]['value']
            elif any(keyword in nameOfValue[s] for keyword in ["AQNTIME", "TIMESTAMP"]):  # search for Interval
                self.Time = content['namesAndValues'][s]['value']
            elif self.SignalName in nameOfValue[s]:  # search for SignalName
                self.Value = content['namesAndValues'][s]['value']
            else:
                raise ValueError("Keyword {} not supported for this type of query.".format(nameOfValue[s]))

        if self.Time == {}:  # Time empty, has to be generated from startTime and intervTime
            self.Time = [t for t in
                         range(self.StartTime, self.StartTime + self.IntervTime * len(self.Value), self.IntervTime)]