import unittest
import pandas as pd
from datetime import datetime
import pytz
from lhcsmapi.Time import Time as tc


class TestTime(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

        # region convertToUnixTime

    def test_convertToUnixTimestamp_realNumberInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToUnixTimestamp returns the correct integer (ns unixTimestamp) when passed real number values"""

        # arrange:
        testCases = \
            [
                (1, None, 1),
                (1, 'ns', 1),
                (1.0, None, 1),
                (1.0, 'ns', 1),
                (1.32, 'ns', 1),
                (1363149019000000001, 'ns', 1363149019000000001),
                (1.528374951891e+18, 'ns', 1528374951891000000),
                (1534598315.006001, 's', 1534598315006001000),

                (1, 'us', int(1e3)),
                (1.0, 'us', int(1e3)),

                (1, 'ms', int(1e6)),
                (1.0, 'ms', int(1e6)),
                (1.32, 'ms', int(1.32e6)),

                (1, 's', int(1e9)),
                (100, 's', int(100e9)),
                (1.0, 's', int(1e9)),
                (1.2, 's', int(1.2e9)),

            ]

        for inputValue, unit, unixTimeComparison in testCases:
            with self.subTest(name=str(inputValue)):
                # act
                unixTime = tc.convertToUnixTimestamp(inputValue, unit=unit, tz='UTC')
                # assert
                self.assertEqual(unixTime, unixTimeComparison)

    def test_convertToUnixTimestamp_stringInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToUnixTimestamp returns the correct integer (ns unixTimestamp) when passed strings"""

        # arrange:
        testCases = \
            [
                # different precisions
                ('1970-01-01 00:00:00.000000001', None, 1),
                ('1970-01-01 00:00:00.000001', None, 1000),
                ('1970-01-01 00:00:00.001', None, int(1e6)),
                ('1970-01-01 00:00:01', None, int(1e9)),
                ('1970-01-01 00:01:00', None, int(60e9)),
                ('1970-01-01 01:00:00', None, int(3600e9)),

                ('2019-06-17 11:41:11.75', None, 1560771671750000000),
                ('2019-06-17 11:41:11.758001', None, 1560771671758001000),
                ('2019-06-17 11:41:11.758001000', None, 1560771671758001000),
                ('2019-06-17 13:41:11.758001000+02:00', None, 1560771671758001000),

                # different decades and utc offsets
                ('1981-06-20 06:59:02.000000004', None, 361868342000000004),
                ('1981-06-20 06:59:02.000000004+00:00', None, 361868342000000004),
                ('1981-06-20 08:59:02.000000004+02:00', None, 361868342000000004),

                ('2013-03-13 04:30:19.000000001', None, 1363149019000000001),
                ('2013-03-13 04:30:19.000000001+00:00', None, 1363149019000000001),
                ('2013-03-13 05:30:19.000000001+01:00', None, 1363149019000000001),

                ('2027-01-01 11:00:01.000010011', None, 1798801201000010011),
                ('2027-01-01 11:00:01.000010011+00:00', None, 1798801201000010011),
                ('2027-01-01 12:00:01.000010011+01:00', None, 1798801201000010011),

            ]

        for inputString, unit, unixTimeToCompare in testCases:
            with self.subTest(name=str(inputString)):
                # act

                unixTimestamp = tc.convertToUnixTimestamp(inputString, unit=unit, tz='UTC')

                # assert
                self.assertEqual(unixTimestamp, unixTimeToCompare)

    def test_convertToUnixTimestamp_datetimeInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToUnixTimestamp returns the correct integer (ns unixTimestamp) when passed datetime"""

        # arrange:
        testCases = \
            [
                # different precisions
                (datetime(1970, 1, 1, 0, 0, 0, 0), 0),
                (datetime(1970, 1, 1, 0, 0, 0, 1), 1000),
                (datetime(1970, 1, 1, 0, 0, 0, 1000), int(1e6)),
                (datetime(1970, 1, 1, 0, 0, 1, 0), int(1e9)),
                (datetime(1970, 1, 1, 0, 1, 0, 0), int(60e9)),
                (datetime(1970, 1, 1, 1, 0, 0, 0), int(3600e9)),
                (datetime(1981, 6, 20, 6, 59, 2, 0), 361868342000000000),
                (datetime(2013, 3, 13, 4, 30, 19, 0), 1363149019000000000),
                (datetime(2019, 6, 17, 11, 41, 11, 750000), 1560771671750000000),
                (datetime(2027, 1, 1, 11, 00, 1, 10), 1798801201000010000)

            ]

        for dtInput, unixTimeToCompare in testCases:
            with self.subTest(name=str(dtInput)):
                # act

                unixTimestamp = tc.convertToUnixTimestamp(dtInput, tz='UTC')

                # assert
                self.assertEqual(unixTimestamp, unixTimeToCompare)

    def test_convertToUnixTimestamp_ZurichTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToUnixTimestamp eturns the correct integer (ns unixTimestamp) when other time zone is used"""

        # arrange:
        testCases = \
            [
                (1, 'ns', 1),
                ('1970-01-01 01:00:00.000000001', None, 1),
                (datetime(1970, 1, 1, 1, 0, 1, 0), None, int(1e9)),
            ]

        for inputValue, unit, unixTimeToCompare in testCases:
            with self.subTest(name=str(inputValue)):
                # act

                unixTimestamp = tc.convertToUnixTimestamp(inputValue, unit=unit, tz='Europe/Zurich')

                # assert
                self.assertEqual(unixTimestamp, unixTimeToCompare)

        # endregion

        # region convertToPandasTimestamp

    def test_convertToPandasTimestamp_realNumberInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToPandasTimestamp returns the correct pandas Timestamp when passed real number values"""

        # arrange:
        testCases = \
            [
                (1, None, 1),
                (1, 'ns', 1),
                (1.0, None, 1),
                (1.0, 'ns', 1),
                (1.32, 'ns', 1),
                (1363149019000000001, 'ns', 1363149019000000001),
                (1.528374951891e+18, 'ns', 1528374951891000000),

                (1, 'us', int(1e3)),
                (1.0, 'us', int(1e3)),

                (1, 'ms', int(1e6)),
                (1.0, 'ms', int(1e6)),
                (1.32, 'ms', int(1.32e6)),

                (1, 's', int(1e9)),
                (100, 's', int(100e9)),
                (1.0, 's', int(1e9)),
                (1.2, 's', int(1.2e9)),
                (1534597650.006001, 's', 1534597650006001000)

            ]

        for inputValue, unit, unixTime in testCases:
            with self.subTest(name=str(inputValue)):
                # act
                pdTimestamp = tc.convertToPandasTimestamp(inputValue, unit=unit, tz='UTC')
                pdTimestampComparison = pd.Timestamp(unixTime, tz='UTC')

                # assert
                self.assertEqual(pdTimestamp, pdTimestampComparison)

    def test_convertToPandasTimestamp_stringInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToPandasTimestamp returns the correct pandas Timestamp when passed strings"""

        # arrange:
        testCases = \
            [
                # different precisions
                ('1970-01-01 00:00:00.000000001', 1),
                ('1970-01-01 00:00:00.000001', 1000),
                ('1970-01-01 00:00:00.001', int(1e6)),
                ('1970-01-01 00:00:01', int(1e9)),
                ('1970-01-01 00:01:00', int(60e9)),
                ('1970-01-01 01:00:00', int(3600e9)),

                ('2019-06-17 11:41:11.75', 1560771671750000000),
                ('2019-06-17 11:41:11.758001', 1560771671758001000),
                ('2019-06-17 11:41:11.758001000', 1560771671758001000),
                ('2019-06-17 13:41:11.758001000+02:00', 1560771671758001000),

                # different decades and utc offsets
                ('1981-06-20 06:59:02.000000004', 361868342000000004),
                ('1981-06-20 06:59:02.000000004+00:00', 361868342000000004),
                ('1981-06-20 08:59:02.000000004+02:00', 361868342000000004),

                ('2013-03-13 04:30:19.000000001', 1363149019000000001),
                ('2013-03-13 04:30:19.000000001+00:00', 1363149019000000001),
                ('2013-03-13 05:30:19.000000001+01:00', 1363149019000000001),

                ('2027-01-01 11:00:01.000010011', 1798801201000010011),
                ('2027-01-01 11:00:01.000010011+00:00', 1798801201000010011),
                ('2027-01-01 12:00:01.000010011+01:00', 1798801201000010011),

            ]

        for inputString, unixTime in testCases:
            with self.subTest(name=str(inputString)):
                # act

                pdTimestamp = tc.convertToPandasTimestamp(inputString, tz='UTC')
                pdTimestampComparison = pd.Timestamp(unixTime, tz='UTC')

                # assert
                self.assertEqual(pdTimestamp, pdTimestampComparison)

    def test_convertToPandasTimestamp_datetimeInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToUnixTimestamp returns the correct integer (ns unixTimestamp) when passed datetime"""

        # arrange:
        testCases = \
            [
                # different precisions
                (datetime(1970, 1, 1, 0, 0, 0, 0), 0),
                (datetime(1970, 1, 1, 0, 0, 0, 1), 1000),
                (datetime(1970, 1, 1, 0, 0, 0, 1000), int(1e6)),
                (datetime(1970, 1, 1, 0, 0, 1, 0), int(1e9)),
                (datetime(1970, 1, 1, 0, 1, 0, 0), int(60e9)),
                (datetime(1970, 1, 1, 1, 0, 0, 0), int(3600e9)),
                (datetime(1981, 6, 20, 6, 59, 2, 0), 361868342000000000),
                (datetime(2013, 3, 13, 4, 30, 19, 0), 1363149019000000000),
                (datetime(2019, 6, 17, 11, 41, 11, 750000), 1560771671750000000),
                (datetime(2027, 1, 1, 11, 00, 1, 10), 1798801201000010000)

            ]

        for dtInput, unixTime in testCases:
            with self.subTest(name=str(dtInput)):
                # act

                pdTimestamp = tc.convertToDatetime(dtInput, tz='UTC')
                pdTimestampComparison = pd.Timestamp(unixTime, tz='UTC')

                # assert
                self.assertEqual(pdTimestamp, pdTimestampComparison)

    def test_convertToPandasTimestamp_ZurichTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToPandasTimestamp returns the correct pandas Timestamp when other time zone is used"""

        # arrange:
        # arrange:
        testCases = \
            [
                (1, 'ns', pd.Timestamp(1, tz='Europe/Zurich')),
                ('1970-01-01 01:00:00.000000001', None, pd.Timestamp(1, tz='Europe/Zurich')),
                (datetime(1970, 1, 1, 1, 0, 1, 0), None, pd.Timestamp(int(1e9), tz='Europe/Zurich')),
            ]

        for inputValue, unit, pdTimestampComparison in testCases:
            with self.subTest(name=str(inputValue)):
                # act

                pdTimestamp = tc.convertToPandasTimestamp(inputValue, unit=unit, tz='Europe/Zurich')

                # assert
                self.assertEqual(pdTimestamp, pdTimestampComparison)

        # endregion

        # region convertString

    def test_convertToString_realNumberInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToString returns the correct string when passed real number values"""

        # arrange:
        testCases = \
            [
                (1, None, '1970-01-01 00:00:00.000000001+00:00'),
                (1, 'ns', '1970-01-01 00:00:00.000000001+00:00'),
                (1.0, None, '1970-01-01 00:00:00.000000001+00:00'),
                (1.0, 'ns', '1970-01-01 00:00:00.000000001+00:00'),
                (1.32, 'ns', '1970-01-01 00:00:00.000000001+00:00'),
                (1363149019000000001, 'ns', '2013-03-13 04:30:19.000000001+00:00'),
                (1.528374951891e+18, 'ns', '2018-06-07 12:35:51.891000+00:00'),

                (1, 'us', '1970-01-01 00:00:00.000001+00:00'),
                (1.0, 'us', '1970-01-01 00:00:00.000001+00:00'),

                (1, 'ms', '1970-01-01 00:00:00.001000+00:00'),
                (1.0, 'ms', '1970-01-01 00:00:00.001000+00:00'),
                (1.32, 'ms', '1970-01-01 00:00:00.001320+00:00'),

                (1, 's', '1970-01-01 00:00:01+00:00'),
                (100, 's', '1970-01-01 00:01:40+00:00'),
                (100.0, 's', '1970-01-01 00:01:40+00:00'),
                (1.0, 's', '1970-01-01 00:00:01+00:00'),
                (1.2, 's', '1970-01-01 00:00:01.200000+00:00'),

            ]

        for inputValue, unit, comparisonString in testCases:
            with self.subTest(name=str(inputValue)):
                # act

                outputString = tc.convertToString(inputValue, unit=unit, tz='UTC')

                # assert
                self.assertEqual(outputString, comparisonString)

    def test_convertToString_stringInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToString returns a correctly formatted string when parsed another string"""
        # arrange:

        testCases = [
            # different precisions
            ('1970-01-01 00:00:00.000000001', '1970-01-01 00:00:00.000000001+00:00'),
            ('1970-01-01 00:00:00.000001', '1970-01-01 00:00:00.000001+00:00'),
            ('1970-01-01 00:00:00.001', '1970-01-01 00:00:00.001000+00:00'),
            ('1970-01-01 00:00:01', '1970-01-01 00:00:01+00:00'),
            ('1970-01-01 00:01:00', '1970-01-01 00:01:00+00:00'),
            ('1970-01-01 01:00:00', '1970-01-01 01:00:00+00:00'),

            ('2019-06-17 11:41:11.75', '2019-06-17 11:41:11.750000+00:00'),
            ('2019-06-17 11:41:11.758001', '2019-06-17 11:41:11.758001+00:00'),
            ('2019-06-17 11:41:11.758001000', '2019-06-17 11:41:11.758001+00:00'),
            ('2019-06-17 13:41:11.758001000+02:00', '2019-06-17 11:41:11.758001+00:00'),

            # different decades and utc offsets
            ('1981-06-20 06:59:02.000000004', '1981-06-20 06:59:02.000000004+00:00'),
            ('1981-06-20 06:59:02.000000004+00:00', '1981-06-20 06:59:02.000000004+00:00'),
            ('1981-06-20 08:59:02.000000004+02:00', '1981-06-20 06:59:02.000000004+00:00'),

            ('2013-03-13 04:30:19.000000001', '2013-03-13 04:30:19.000000001+00:00'),
            ('2013-03-13 04:30:19.000000001+00:00', '2013-03-13 04:30:19.000000001+00:00'),
            ('2013-03-13 05:30:19.000000001+01:00', '2013-03-13 04:30:19.000000001+00:00'),

            ('2027-01-01 11:00:01.000010011', '2027-01-01 11:00:01.000010011+00:00'),
            ('2027-01-01 11:00:01.000010011+00:00', '2027-01-01 11:00:01.000010011+00:00'),
            ('2027-01-01 12:00:01.000010011+01:00', '2027-01-01 11:00:01.000010011+00:00')]

        for inputString, comparisonString in testCases:
            with self.subTest(name=str(inputString)):
                # act

                outputString = tc.convertToString(inputString, tz='UTC')

                # assert
                self.assertEqual(outputString, comparisonString)

    def test_convertToString_datetimeInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToString returns the correct string when passed datetime"""

        # arrange:
        testCases = \
            [
                # different precisions
                ('1970-01-01 00:00:00+00:00', datetime(1970, 1, 1, 0, 0, 0, 0)),
                ('1970-01-01 00:00:00.000001+00:00', datetime(1970, 1, 1, 0, 0, 0, 1)),
                ('1970-01-01 00:00:00.001000+00:00', datetime(1970, 1, 1, 0, 0, 0, 1000)),
                ('1970-01-01 00:00:01+00:00', datetime(1970, 1, 1, 0, 0, 1, 0)),
                ('1970-01-01 00:01:00+00:00', datetime(1970, 1, 1, 0, 1, 0, 0)),
                ('1970-01-01 01:00:00+00:00', datetime(1970, 1, 1, 1, 0, 0, 0)),

                ('2019-06-17 11:41:11.750000+00:00', datetime(2019, 6, 17, 11, 41, 11, 750000)),
                ('2019-06-17 11:41:11.758001+00:00', datetime(2019, 6, 17, 11, 41, 11, 758001)),

                # different decades and utc offsets
                ('1981-06-20 06:59:02+00:00', datetime(1981, 6, 20, 6, 59, 2, 0)),
                ('2013-03-13 04:30:19+00:00', datetime(2013, 3, 13, 4, 30, 19, 0)),
                ('2027-01-01 11:00:01.000010+00:00', datetime(2027, 1, 1, 11, 0, 1, 10)),

            ]

        for comparisonString, dt in testCases:
            with self.subTest(name=str(dt)):
                # act

                outputString = tc.convertToString(dt, tz='UTC')

                # assert
                self.assertEqual(outputString, comparisonString)

    def test_convertToString_ZurichTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToString returns the correct timestamp string when time zone is used"""

        # arrange:
        # arrange:
        testCases = \
            [
                (1, 'ns', '1970-01-01 01:00:00.000000001+01:00'),
                ('1970-01-01 01:00:00.000000001', None, '1970-01-01 01:00:00.000000001+01:00'),
                (datetime(1970, 1, 1, 1, 0, 1, 0), None, '1970-01-01 01:00:01+01:00'),
            ]

        for inputValue, unit, timestampStringComparison in testCases:
            with self.subTest(name=str(inputValue)):
                # act
                timestampString = tc.convertToString(inputValue, unit=unit, tz='Europe/Zurich')

                # assert
                self.assertEqual(timestampString, timestampStringComparison)

        # endregion

        # region convertToDateTime

    def test_convertToDateTime_realNumberInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToDatetime returns the correct datetime when passed real number values"""

        # arrange:
        testCases = \
            [
                (1, None, datetime(1970, 1, 1, 0, 0, 0, 0)),
                (1, 'ns', datetime(1970, 1, 1, 0, 0, 0, 0)),
                (1.0, None, datetime(1970, 1, 1, 0, 0, 0, 0)),
                (1.0, 'ns', datetime(1970, 1, 1, 0, 0, 0, 0)),
                (1.32, 'ns', datetime(1970, 1, 1, 0, 0, 0, 0)),
                (1363149019000000001, 'ns', datetime(2013, 3, 13, 4, 30, 19, 0)),
                (1.528374951891e+18, 'ns', datetime(2018, 6, 7, 12, 35, 51, 891000)),

                (1, 'us', datetime(1970, 1, 1, 0, 0, 0, 1)),
                (1.0, 'us', datetime(1970, 1, 1, 0, 0, 0, 1)),

                (1, 'ms', datetime(1970, 1, 1, 0, 0, 0, 1000)),
                (1.0, 'ms', datetime(1970, 1, 1, 0, 0, 0, 1000)),
                (1.32, 'ms', datetime(1970, 1, 1, 0, 0, 0, 1320)),

                (1, 's', datetime(1970, 1, 1, 0, 0, 1, 0)),
                (100, 's', datetime(1970, 1, 1, 0, 1, 40, 0)),
                (1.0, 's', datetime(1970, 1, 1, 0, 0, 1, 0)),
                (1.2, 's', datetime(1970, 1, 1, 0, 0, 1, 200000)),

            ]

        for inputValue, unit, dtComparison in testCases:
            with self.subTest(name=str(inputValue)):
                # act
                dt = tc.convertToDatetime(inputValue, unit=unit, tz='UTC')
                dtComparison = dtComparison.replace(tzinfo=pytz.utc)

                # assert
                self.assertEqual(dt, dtComparison)

    def test_convertToDatetime_stringInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToDatetime returns the correct datetime when passed strings"""

        # arrange:
        testCases = \
            [
                # different precisions
                ('1970-01-01 00:00:00.000000001', datetime(1970, 1, 1, 0, 0, 0, 0)),
                ('1970-01-01 00:00:00.000001', datetime(1970, 1, 1, 0, 0, 0, 1)),
                ('1970-01-01 00:00:00.001', datetime(1970, 1, 1, 0, 0, 0, 1000)),
                ('1970-01-01 00:00:01', datetime(1970, 1, 1, 0, 0, 1, 0)),
                ('1970-01-01 00:01:00', datetime(1970, 1, 1, 0, 1, 0, 0)),
                ('1970-01-01 01:00:00', datetime(1970, 1, 1, 1, 0, 0, 0)),

                ('2019-06-17 11:41:11.75', datetime(2019, 6, 17, 11, 41, 11, 750000)),
                ('2019-06-17 11:41:11.758001', datetime(2019, 6, 17, 11, 41, 11, 758001)),
                ('2019-06-17 11:41:11.758001000', datetime(2019, 6, 17, 11, 41, 11, 758001)),
                ('2019-06-17 13:41:11.758001000+02:00', datetime(2019, 6, 17, 11, 41, 11, 758001)),

                # different decades and utc offsets
                ('1981-06-20 06:59:02.000000004', datetime(1981, 6, 20, 6, 59, 2, 0)),
                ('1981-06-20 06:59:02.000000004+00:00', datetime(1981, 6, 20, 6, 59, 2, 0)),
                ('1981-06-20 08:59:02.000000004+02:00', datetime(1981, 6, 20, 6, 59, 2, 0)),

                ('2013-03-13 04:30:19.000000001', datetime(2013, 3, 13, 4, 30, 19, 0)),
                ('2013-03-13 04:30:19.000000001+00:00', datetime(2013, 3, 13, 4, 30, 19, 0)),
                ('2013-03-13 05:30:19.000000001+01:00', datetime(2013, 3, 13, 4, 30, 19, 0)),

                ('2027-01-01 11:00:01.000010011', datetime(2027, 1, 1, 11, 00, 1, 10)),
                ('2027-01-01 11:00:01.000010011+00:00', datetime(2027, 1, 1, 11, 00, 1, 10)),
                ('2027-01-01 12:00:01.000010011+01:00', datetime(2027, 1, 1, 11, 00, 1, 10)),

            ]

        for inputString, dtComparison in testCases:
            with self.subTest(name=str(inputString)):
                # act
                dt = tc.convertToDatetime(inputString, tz='UTC')
                dtComparison = dtComparison.replace(tzinfo=pytz.utc)

                # assert
                self.assertEqual(dt, dtComparison)

    def test_convertToDatetime_datetimeInput_utcTimeZone_returnsCorrectTimestamp(self):
        """Tests that convertToDatetime returns the correct datetime when passed datetimes"""

        # arrange:
        testCases = \
            [
                # different precisions
                datetime(1970, 1, 1, 0, 0, 0, 0),
                datetime(1970, 1, 1, 0, 0, 0, 1),
                datetime(1970, 1, 1, 0, 0, 0, 1000),
                datetime(1970, 1, 1, 0, 0, 1, 0),
                datetime(1970, 1, 1, 0, 1, 0, 0),
                datetime(1970, 1, 1, 1, 0, 0, 0),

                datetime(2019, 6, 17, 11, 41, 11, 750000),
                datetime(2019, 6, 17, 11, 41, 11, 758001),

                # different decades and utc offsets
                datetime(1981, 6, 20, 6, 59, 2, 0),
                datetime(2013, 3, 13, 4, 30, 19, 0),
                datetime(2027, 1, 1, 11, 00, 1, 10),

            ]

        for dtComparison in testCases:
            with self.subTest(name=str(dtComparison)):
                # act
                dt = tc.convertToDatetime(dtComparison, tz='UTC')
                dtComparison = dtComparison.replace(tzinfo=pytz.utc)

                # assert
                self.assertEqual(dt, dtComparison)

        # endregion

        # region getQueryPeriodInUnixTime tests

    def test_getQueryPeriodInUnixTime_startTimeAndEndTimeProvided_tzSetToZurich_covertsToUnixTime(self):
        """Tests that getQueryPeriodInUnixTime returns the right start and endTime if they are both provided. Essentially
        just converts both to unix time"""

        # arrange:
        testCases = \
            [
                ('2018-08-18 15:07:30.006000', '2018-08-18 15:30:32.006000', 1534597650006000000, 1534599032006000000),

            ]

        for startTimeData, endTimeData, refStartTime, refEndTime in testCases:
            with self.subTest():
                # act
                startTime, endTime = tc.getQueryPeriodInUnixTime(startTimeData, endTimeData, None, tz='Europe/Zurich')

                # assert
                self.assertEqual(startTime, refStartTime)
                self.assertEqual(endTime, refEndTime)

    def test_getQueryPeriodInUnixTime_singleDurationTupleProvided_tzSetToZurich_calculatesEndTime(self):
        """Tests that getQueryPeriodInUnixTime returns the right start and endTime if startTime and a single duration component is provided."""

        # arrange:
        testCases = \
            [
                ('2018-08-18 15:07:30.006000', (2, 's'), 1534597650006000000, 1534597652006000000),
                ('2018-08-18 15:07:30.006000', (60 * 30, 's'), 1534597650006000000, 1534599450006000000)

            ]

        for startTimeData, durationData, refStartTime, refEndTime in testCases:
            with self.subTest():
                # act
                startTime, endTime = tc.getQueryPeriodInUnixTime(startTimeData, None, durationData, tz='Europe/Zurich')

                # assert
                self.assertEqual(startTime, refStartTime)
                self.assertEqual(endTime, refEndTime)

    def test_getQueryPeriodInUnixTime_singleDurationTupleProvidedInList_tzSetToZurich_calculatesEndTime(self):
        """Tests that getQueryPeriodInUnixTime returns the right start and endTime if startTime and a single duration component is provided
        in list format."""

        # arrange:
        testCases = \
            [
                ('2018-08-18 15:07:30.006000', [(2, 's')], 1534597650006000000, 1534597652006000000),
                ('2018-08-18 15:07:30.006000', [(60 * 30, 's')], 1534597650006000000, 1534599450006000000)

            ]

        for startTimeData, durationData, refStartTime, refEndTime in testCases:
            with self.subTest():
                # act
                startTime, endTime = tc.getQueryPeriodInUnixTime(startTimeData, None, durationData, tz='Europe/Zurich')

                # assert
                self.assertEqual(startTime, refStartTime)
                self.assertEqual(endTime, refEndTime)

    def test_getQueryPeriodInUnixTime_TwoDurationTupleProvidedInList_tzSetToZurich_calculatesBothStartAndEndTime(self):
        """Tests that getQueryPeriodInUnixTime returns the right start and endTime if startTime and two duration components are provided
        in list format."""

        # arrange:
        testCases = \
            [
                ('2018-08-18 15:07:30.006000', [(2, 's'), (2, 's')], 1534597648006000000, 1534597652006000000),
                ('2018-08-18 15:07:30.006000', [(60 * 30, 's'), (60 * 30, 's')], 1534595850006000000,
                 1534599450006000000)

            ]

        for startTimeData, durationData, refStartTime, refEndTime in testCases:
            with self.subTest():
                # act
                startTime, endTime = tc.getQueryPeriodInUnixTime(startTimeData, None, durationData, tz='Europe/Zurich')

                # assert
                self.assertEqual(startTime, refStartTime)
                self.assertEqual(endTime, refEndTime)

    def test_moduloDayInUnixTimestamp_error(self):
        # arrange
        startTime = '2018-12-13 00:00:00'
        endTime = '2018-12-12 00:00:00'

        # act

        # assert
        with self.assertRaises(AttributeError):
            tc.moduloDayInUnixTimestamp(startTime, endTime)

    def test_moduloDayInUnixTimestamp_less_than_a_day(self):
        # arrange
        startTime = '2018-12-12 00:00:00'
        endTime = '2018-12-12 01:00:00'

        # act
        quotientAct, remainderAct = tc.moduloDayInUnixTimestamp(startTime, endTime)

        # assert
        quotientExp = 0
        remainderExp = int(1*60*60*1e9)
        self.assertEqual(quotientExp, quotientAct)
        self.assertEqual(remainderAct, remainderExp)

    def test_moduloDayInUnixTimestamp_single_day(self):
        # arrange
        startTime = '2018-12-12 00:00:00'
        endTime = '2018-12-13 00:00:00'

        # act
        quotientAct, remainderAct = tc.moduloDayInUnixTimestamp(startTime, endTime)

        # assert
        quotientExp = 1
        remainderExp = 0
        self.assertEqual(quotientExp, quotientAct)
        self.assertEqual(remainderAct, remainderExp)

    def test_moduloDayInUnixTimestamp_two_days(self):
        # arrange
        startTime = '2018-12-12 00:00:00'
        endTime = '2018-12-14 00:00:00'

        # act
        quotientAct, remainderAct = tc.moduloDayInUnixTimestamp(startTime, endTime)

        # assert
        quotientExp = 2
        remainderExp = 0
        self.assertEqual(quotientExp, quotientAct)
        self.assertEqual(remainderAct, remainderExp)

    def test_moduloDayInUnixTimestamp_two_days_and_a_half(self):
        # arrange
        startTime = '2018-12-12 00:00:00'
        endTime = '2018-12-14 12:00:00'

        # act
        quotientAct, remainderAct = tc.moduloDayInUnixTimestamp(startTime, endTime)

        # assert
        quotientExp = 2
        remainderExp = int(12*60*60*1e9)
        self.assertEqual(quotientExp, quotientAct)
        self.assertEqual(remainderAct, remainderExp)

    def test_timezone_conversion(self):
        # arrange
        dateExp = "2018-12-12 00:00:00+01:00"

        # act
        timestamp = tc.convertToUnixTimestamp(dateExp)
        dateAct = tc.convertToString(timestamp)
        self.assertEqual(dateExp, dateAct)



