import math
import unittest
import pandas as pd
from pathlib import Path
import os
import numpy as np

from lhcsmapi.Time import Time
from lhcsmapi.analysis.DiodeLeadResistanceAnalysis import DiodeLeadResistanceAnalysis
from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


class TestDiodeLeadResistanceAnalysis(unittest.TestCase):

    @staticmethod
    def parseCalsSignal(signalName, magnetName):
        path = Path(os.path.dirname(__file__))

        if 'I_MEAS' in signalName:
            filePath = os.path.join(str(path.parent), '../test/resources/diode_lead_resistance/cals/I_MEAS_{}.csv'.format(magnetName))
        elif 'U_DIODE' in signalName:
            filePath = os.path.join(str(path.parent), '../test/resources/diode_lead_resistance/cals/U_DIODE_{}.csv'.format(magnetName))
        else:
            raise KeyError('signalName {} not supported. Suported signalName are {}'.format(signalName, ['I_MEAS', 'U_DIODE_%']))

        return pd.read_csv(filePath, index_col=0)

    @staticmethod
    def parseCalsSyncSignal(signalName, magnetName):
        path = Path(os.path.dirname(__file__))

        if signalName == 'RPHE.UA47.RQD.A45:I_MEAS':
            filePath = os.path.join(str(path.parent), '../test/resources/diode_lead_resistance/cals/I_MEAS_SYNC_{}.csv'.format(magnetName))
        elif signalName == 'MQ.25R4:U_DIODE_RQD':
            filePath = os.path.join(str(path.parent), '../test/resources/diode_lead_resistance/cals/U_DIODE_SYNC_{}.csv'.format(magnetName))
        else:
            raise KeyError('signalName {} not supported. Suported signalName are {}'.format(signalName, ['RPHE.UA47.RQD.A45:I_MEAS', 'MQ.25R4:U_DIODE_RQD']))

        return pd.read_csv(filePath, index_col=0)

    @staticmethod
    def parseAnalysisOutput(signalName, magnetName):
        path = Path(os.path.dirname(__file__))

        if signalName == 'RPHE.UA47.RQD.A45:I_MEAS':
            filePath = os.path.join(str(path.parent), '../test/resources/diode_lead_resistance/cals/I_MEAS_OUTPUT_{}.csv'.format(magnetName))
        elif signalName == 'MQ.25R4:U_DIODE_RQD':
            filePath = os.path.join(str(path.parent), '../test/resources/diode_lead_resistance/cals/U_DIODE_OUTPUT_{}.csv'.format(magnetName))
        else:
            raise KeyError('signalName {} not supported. Suported signalName are {}'.format(signalName,
                                                                                            ['RPHE.UA47.RQD.A45:I_MEAS',
                                                                                             'MQ.25R4:U_DIODE_RQD']))

        return pd.read_csv(filePath, index_col=0)

    @staticmethod
    def parseCalsSyncResampleSignal(signalName, magnetName):
        path = Path(os.path.dirname(__file__))

        if signalName == 'RPHE.UA47.RQD.A45:I_MEAS':
            filePath = os.path.join(str(path.parent), '../test/resources/diode_lead_resistance/cals/I_MEAS_SYNC_RESAMPLE_{}.csv'.format(magnetName))
        else:
            raise KeyError('signalName {} not supported. Suported signalName is \'RPHE.UA47.RQD.A45:I_MEAS\'' )

        return pd.read_csv(filePath, index_col=0)

    @staticmethod
    def parsePmSignal(signalName):
        path = Path(os.path.dirname(__file__))
        try:
            filePath = os.path.join(str(path.parent),
                                    '../test/resources/diode_lead_resistance/pm/{}.csv'.format(signalName))
            return pd.read_csv(filePath, index_col=0)
        except:
            raise KeyError('signalName {} not supported'.format(signalName))

    def test_i_meas_name_rqd(self):
        # arrange
        circuitType = 'RQD'
        circuitName = 'RQD.A45'
        systemType = 'PC'
        databaseName = 'CALS'
        variableName = 'I_MEAS'

        # act
        i_meas_name_act = SignalMetadata.getVariableName(circuitType, circuitName, systemType, databaseName, variableName)

        # assert
        i_meas_name_exp = 'RPHE.UA47.RQD.A45:I_MEAS'
        self.assertEqual(i_meas_name_exp, i_meas_name_act)

    def test_u_diode_name_rqd(self):
        # arrange
        circuitType = 'RQD'
        circuitName = 'RQD.A45'
        magnetName = 'MQ.25R4'
        systemType = 'DIODE_RQD'
        databaseName = 'CALS'
        variableName = 'U_DIODE_RQD'

        # act
        u_diode_name_act = SignalMetadata.getVariableName(circuitType, circuitName, systemType, databaseName, variableName)
        u_diode_name_act = u_diode_name_act.replace("%MAGNET%", magnetName)

        # assert
        u_diode_name_exp = 'MQ.25R4:U_DIODE_RQD'
        self.assertEqual(u_diode_name_exp, u_diode_name_act)

    def test_u_diode_name_rb(self):
        # arrange
        circuitType = 'RB'
        circuitName = 'RB.A12'
        magnetName = 'MB.A31R1'
        systemType = 'DIODE_RB'
        databaseName = 'CALS'
        variableName = 'U_DIODE_RB'

        # act
        u_diode_name_act = SignalMetadata.getVariableName(circuitType, circuitName, systemType, databaseName, variableName)
        u_diode_name_act = u_diode_name_act.replace("%MAGNET%", magnetName)

        # assert
        u_diode_name_exp = 'MB.A31R1:U_DIODE_RB'
        self.assertEqual(u_diode_name_exp, u_diode_name_act)

    def test_i_meas_sync_values(self):
        # arrange
        signalName = 'RPHE.UA47.RQD.A45:I_MEAS'
        timestamp = '2018-12-09 09:55:47.668'
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)
        magnetName = 'MQ.25R4'

        # act
        i_meas = self.parseCalsSignal(signalName, magnetName)
        i_meas_sync_act = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)

        # assert
        i_meas_sync_exp = self.parseCalsSyncSignal(signalName, magnetName)
        self.assertListEqual(list(i_meas_sync_exp[signalName].values), list(i_meas_sync_act[signalName].values))

    def test_i_meas_sync_index(self):
        # arrange
        signalName = 'RPHE.UA47.RQD.A45:I_MEAS'
        timestamp = '2018-12-09 09:55:47.668'
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)
        magnetName = 'MQ.25R4'

        # act
        i_meas = self.parseCalsSignal(signalName, magnetName)
        i_meas_sync_act = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)

        # assert
        i_meas_sync_exp = self.parseCalsSyncSignal(signalName, magnetName)
        for index_exp, index_act in zip(list(i_meas_sync_exp[signalName].index), list(i_meas_sync_act[signalName].index)):
            self.assertAlmostEqual(index_exp, index_act, places=6)

    def test_u_diode_sync_values(self):
        # arrange
        signalName = 'MQ.25R4:U_DIODE_RQD'
        timestamp = '2018-12-09 09:55:47.668'
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)
        magnetName = 'MQ.25R4'

        # act
        u_diode = self.parseCalsSignal(signalName, magnetName)
        u_diode_sync_act = SignalUtilities.synchronizeDataframe(u_diode, timestampUnixSec)

        # assert
        u_diode_sync_exp = self.parseCalsSyncSignal(signalName, magnetName)
        self.assertListEqual(list(u_diode_sync_exp[signalName].values), list(u_diode_sync_act[signalName].values))

    def test_u_diode_sync_index(self):
        # arrange
        signalName = 'MQ.25R4:U_DIODE_RQD'
        timestamp = '2018-12-09 09:55:47.668'
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)
        magnetName = 'MQ.25R4'

        # act
        i_meas = self.parseCalsSignal(signalName, magnetName)
        i_meas_sync_act = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)

        # assert
        i_meas_sync_ext = self.parseCalsSyncSignal(signalName, magnetName)
        for index_exp, index_act in zip(list(i_meas_sync_ext[signalName].index), list(i_meas_sync_act[signalName].index)):
            self.assertAlmostEqual(index_exp, index_act, places=6)

    def test_i_meas_sync_resample_index(self):
        # arrange
        signalName = 'RPHE.UA47.RQD.A45:I_MEAS'
        timestamp = '2018-12-09 09:55:47.668'
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)
        magnetName = 'MQ.25R4'

        # act
        i_meas = self.parseCalsSignal(signalName, magnetName)
        i_meas_sync = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)
        i_meas_sync_resample_act = SignalAnalysis.resampleSeries(i_meas_sync, ts=0.1, t0=-50, t_end=150)

        # assert
        i_meas_sync_resample_exp = self.parseCalsSyncResampleSignal(signalName, magnetName)
        for index_exp, index_act in zip(list(i_meas_sync_resample_exp[signalName].index), list(i_meas_sync_resample_act[signalName].index)):
            self.assertAlmostEqual(index_exp, index_act, places=10)

    def test_i_meas_sync_resample_values(self):
        # arrange
        signalName = 'RPHE.UA47.RQD.A45:I_MEAS'
        timestamp = '2018-12-09 09:55:47.668'
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)
        magnetName = 'MQ.25R4'

        # act
        i_meas = self.parseCalsSignal(signalName, magnetName)
        i_meas_sync = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)
        i_meas_sync_resample_act = SignalAnalysis.resampleSeries(i_meas_sync, ts=0.1, t0=-50, t_end=150)

        # assert
        i_meas_sync_resample_exp = self.parseCalsSyncResampleSignal(signalName, magnetName)
        for index_exp, index_act in zip(list(i_meas_sync_resample_exp[signalName].values), list(i_meas_sync_resample_act[signalName].values)):
            if math.isnan(index_exp):
                pass
            else:
                self.assertAlmostEqual(index_exp, index_act, places=3)

    def test_u_diode_sync_resample_index(self):
        '''
        The goal of this test is to compare the resampled indices of circuit current and diode voltage.
        This is needed in order to concatenate resampled signals.
        '''
        # arrange
        signalName = 'MQ.25R4:U_DIODE_RQD'
        timestamp = '2018-12-09 09:55:47.668'
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)
        magnetName = 'MQ.25R4'

        # act
        i_meas = self.parseCalsSignal(signalName, magnetName)
        i_meas_sync = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)
        i_meas_sync_resample_act = SignalAnalysis.resampleSeries(i_meas_sync, ts=0.1, t0=-50, t_end=150)

        # assert
        i_meas_sync_resample_exp = self.parseCalsSyncResampleSignal('RPHE.UA47.RQD.A45:I_MEAS', magnetName)
        for index_exp, index_act in zip(list(i_meas_sync_resample_exp['RPHE.UA47.RQD.A45:I_MEAS'].index), list(i_meas_sync_resample_act[signalName].index)):
            self.assertAlmostEqual(index_exp, index_act, places=10)

    def test_u_res_calculation(self):
        # arrange
        U_DIODE = 'MQ.25R4:U_DIODE_RQD'
        magnetName = 'MQ.25R4'

        # act
        uDiode = self.parseCalsSyncSignal(U_DIODE, magnetName)
        uD = uDiode[U_DIODE]
        uRes = []
        # "5" - toggling period of nQPS U_DIODE signals
        # "50s x 10Hz + 20" delay to the diode full opening
        for i in range(0, uD.size):
            if i < 520 or i > uD.size - 6:
                uRes.append(0)
            else:
                uRes.append(abs(uD.iloc[i] - (uD.iloc[i - 5] + uD.iloc[i + 5]) / 2))

        # assert
        uD_exp = self.parseAnalysisOutput(U_DIODE, magnetName)
        uRes_exp = list(uD_exp['uRes'].values)
        for u_res_exp, u_res_act in zip(uRes_exp, uRes):
            self.assertAlmostEqual(u_res_exp, u_res_act, places=10)

    def test_u_res_calculation_rolling(self):
        # arrange
        U_DIODE = 'MQ.25R4:U_DIODE_RQD'
        magnetName = 'MQ.25R4'

        # act
        uDiode = self.parseCalsSyncSignal(U_DIODE, magnetName)
        calc_u_res = lambda x: abs(x[5] - (x[0] + x[-1]) / 2)
        uDiode['uRes2'] = uDiode['MQ.25R4:U_DIODE_RQD'].rolling(11).apply(calc_u_res)
        uDiode.loc[uDiode.index < uDiode.index[520], 'uRes2'] = 0
        uRes_act = list(uDiode['uRes2'].values)

        # assert
        uD_exp = self.parseAnalysisOutput(U_DIODE, magnetName)
        uRes_exp = list(uD_exp['uRes'].values)
        for u_res_exp, u_res_act in zip(uRes_exp, uRes_act):
            self.assertAlmostEqual(u_res_exp, u_res_act, places=0)


    def test_u_d_min_calculation(self):
        # arrange
        U_DIODE = 'MQ.25R4:U_DIODE_RQD'
        magnetName = 'MQ.25R4'

        # act
        uDiode = self.parseCalsSyncSignal(U_DIODE, magnetName)
        uD = uDiode[U_DIODE]
        uD_min = []
        # "5" - toggling period of nQPS U_DIODE signals
        # "50s x 10Hz + 20" delay to the diode full opening
        for i in range(0, uD.size):
            if i < 520 or i > uD.size - 6:
                uD_min.append(0)
            else:
                uD_min.append(min((uD.iloc[i - 5] + uD.iloc[i + 5]) / 2, uD.iloc[i]))

        # assert
        uD_exp = self.parseAnalysisOutput(U_DIODE, magnetName)
        uDmin_exp = list(uD_exp['uD_min'].values)
        for u_d_min_exp, u_d_min_act in zip(uDmin_exp, uD_min):
            self.assertAlmostEqual(u_d_min_exp, u_d_min_act, places=10)

    def test_u_d_min_calculation_rolling(self):
        # arrange
        U_DIODE = 'MQ.25R4:U_DIODE_RQD'
        magnetName = 'MQ.25R4'

        # act
        uDiode = self.parseCalsSyncSignal(U_DIODE, magnetName)
        calc_u_diode_min = lambda x: min((x[0] + x[-1]) / 2, x[5])
        uDiode['uD_min2'] = uDiode['MQ.25R4:U_DIODE_RQD'].rolling(11).apply(calc_u_diode_min)
        uDiode.loc[uDiode.index < uDiode.index[520], 'uD_min2'] = 0
        uDmin_act = list(uDiode['uD_min2'].values)

        # assert
        uD_exp = self.parseAnalysisOutput(U_DIODE, magnetName)
        uDmin_exp = list(uD_exp['uD_min'].values)
        for u_d_min_exp, u_d_min_act in zip(uDmin_exp, uDmin_act):
            self.assertAlmostEqual(u_d_min_exp, u_d_min_act, places=0)

    def test_r_max_t_r_max_i_r_max_MQ25R4(self):
        # arrange
        magnet = '25R4'
        circuitName = 'RQD.A45'
        timestamp = '2018-12-09 09:55:47.668'
        magnetName = 'MQ.25R4'

        # act
        # get metadata - signal names
        magnetType, I_MEAS, U_DIODE = DiodeLeadResistanceAnalysis.get_magnet_type_i_meas_u_diode(magnet, circuitName)

        tEnd = 350 if magnetType == "MB" else 150
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)

        # query circuit current
        i_meas = self.parseCalsSignal(I_MEAS, magnetName)
        i_meas_sync = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)
        i_meas_sync_resampled = SignalAnalysis.resampleSeries(i_meas_sync, ts=0.1, t0=-50, t_end=tEnd)

        # query diode voltage - not resampled
        u_diode = self.parseCalsSignal(U_DIODE, magnetName)
        u_diode_sync = SignalUtilities.synchronizeDataframe(u_diode, timestampUnixSec)

        # join current and diode voltage into a single dataframe
        u_diode_sync[I_MEAS] = i_meas_sync_resampled[I_MEAS].values

        # recover resistive and diode minimum voltage
        DiodeLeadResistanceAnalysis.calculate_minimum_diode_and_current_leads_voltage(u_diode_sync, U_DIODE)

        # calculate resistance
        u_diode_sync['RES'] = u_diode_sync.apply(lambda row: 1e+6 * row['U_RES'] / row[I_MEAS], axis=1, raw=False)

        # set to 0 resistance values for too low current and with too high diode voltage derivative
        DiodeLeadResistanceAnalysis.set_to_zero_resistance_outside_of_range(u_diode_sync, I_MEAS)

        # filter resistance
        DiodeLeadResistanceAnalysis.filter_resistance(u_diode_sync)

        resMaxAct, tResMaxAct, iResMaxAct = DiodeLeadResistanceAnalysis.extract_res_max_time_current_cals(u_diode_sync, I_MEAS)

        # assert
        resMaxExp = 294.8703042295249
        tResMaxExp = 84.67599987983704
        iResMaxExp = 486.2337503851054
        self.assertAlmostEqual(resMaxExp, resMaxAct, places=6)
        self.assertAlmostEqual(tResMaxExp, tResMaxAct, places=6)
        self.assertAlmostEqual(iResMaxExp, iResMaxAct, places=6)

    def test_r_max_t_r_max_i_r_max_MQ30R8(self):
        # arrange
        magnet = '30R8'
        circuitName = 'RQD.A81'
        timestamp = '2018-12-03 15:21:05.372'
        magnetName = 'MQ.30R8'

        # act
        # get metadata - signal names
        magnetType, I_MEAS, U_DIODE = DiodeLeadResistanceAnalysis.get_magnet_type_i_meas_u_diode(magnet, circuitName)

        tEnd = 350 if magnetType == "MB" else 150
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)

        # query circuit current
        i_meas = self.parseCalsSignal(I_MEAS, magnetName)
        i_meas_sync = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)
        i_meas_sync_resampled = SignalAnalysis.resampleSeries(i_meas_sync, ts=0.1, t0=-50, t_end=tEnd)

        # query diode voltage - not resampled
        u_diode = self.parseCalsSignal(U_DIODE, magnetName)
        u_diode_sync = SignalUtilities.synchronizeDataframe(u_diode, timestampUnixSec)

        # join current and diode voltage into a single dataframe
        u_diode_sync[I_MEAS] = i_meas_sync_resampled[I_MEAS].values

        # recover resistive and diode minimum voltage
        DiodeLeadResistanceAnalysis.calculate_minimum_diode_and_current_leads_voltage(u_diode_sync, U_DIODE)

        # calculate resistance
        u_diode_sync['RES'] = u_diode_sync.apply(lambda row: 1e+6 * row['U_RES'] / row[I_MEAS], axis=1, raw=False)

        # set to 0 resistance values for too low current and with too high diode voltage derivative
        DiodeLeadResistanceAnalysis.set_to_zero_resistance_outside_of_range(u_diode_sync, I_MEAS)

        # filter resistance
        DiodeLeadResistanceAnalysis.filter_resistance(u_diode_sync)

        resMaxAct, tResMaxAct, iResMaxAct = DiodeLeadResistanceAnalysis.extract_res_max_time_current_cals(u_diode_sync, I_MEAS)

        # assert
        resMaxExp = 150.24575385608296
        tResMaxExp = 78.67000007629395
        iResMaxExp = 580.12099546584
        self.assertAlmostEqual(resMaxExp, resMaxAct, places=6)
        self.assertAlmostEqual(tResMaxExp, tResMaxAct, places=6)
        self.assertAlmostEqual(iResMaxExp, iResMaxAct, places=6)

    def test_r_max_t_r_max_i_r_max_MB31R1(self):
        # arrange
        magnet = 'A31R1'
        circuitName = 'RB.A12'
        timestamp = '2018-12-12 17:21:34.793'
        magnetName = 'MB.A31R1'

        # act
        # get metadata - signal names
        magnetType, I_MEAS, U_DIODE = DiodeLeadResistanceAnalysis.get_magnet_type_i_meas_u_diode(magnet, circuitName)

        tEnd = 350 if magnetType == "MB" else 150
        timestampUnixSec = Time.convertToUnixTimeSeconds(timestamp)

        # query circuit current
        i_meas = self.parseCalsSignal(I_MEAS, magnetName)
        i_meas_sync = SignalUtilities.synchronizeDataframe(i_meas, timestampUnixSec)
        i_meas_sync_resampled = SignalAnalysis.resampleSeries(i_meas_sync, ts=0.1, t0=-50, t_end=tEnd)

        # query diode voltage - not resampled
        u_diode = self.parseCalsSignal(U_DIODE, magnetName)
        u_diode_sync = SignalUtilities.synchronizeDataframe(u_diode, timestampUnixSec)

        # join current and diode voltage into a single dataframe
        u_diode_sync[I_MEAS] = i_meas_sync_resampled[I_MEAS].values

        # recover resistive and diode minimum voltage
        DiodeLeadResistanceAnalysis.calculate_minimum_diode_and_current_leads_voltage(u_diode_sync, U_DIODE)

        # calculate resistance
        u_diode_sync['RES'] = u_diode_sync.apply(lambda row: 1e+6 * row['U_RES'] / row[I_MEAS], axis=1, raw=False)

        # set to 0 resistance values for too low current and with too high diode voltage derivative
        DiodeLeadResistanceAnalysis.set_to_zero_resistance_outside_of_range(u_diode_sync, I_MEAS)

        # filter resistance
        DiodeLeadResistanceAnalysis.filter_resistance(u_diode_sync)

        resMaxAct, tResMaxAct, iResMaxAct = DiodeLeadResistanceAnalysis.extract_res_max_time_current_cals(u_diode_sync, I_MEAS)

        # assert
        resMaxExp = 31.72004291726936
        tResMaxExp = 4.148000001907349
        iResMaxExp = 11018.072893153274
        self.assertAlmostEqual(resMaxExp, resMaxAct, places=6)
        self.assertAlmostEqual(tResMaxExp, tResMaxAct, places=6)
        self.assertAlmostEqual(iResMaxExp, iResMaxAct, places=6)

    def test_U_DIODE_RQD_A(self):
        # arrange
        u_diode_rqd_a = self.parsePmSignal('U_DIODE_RQD_A')
        u_diode_rqd_a_name = 'MQ.25R4:U_DIODE_RQD'
        timestamp_iqps_qh = '2018-12-09 09:55:47.668'
        timestamp_unix = Time.convertToUnixTimestamp(timestamp_iqps_qh)
        time_corr = 0.0

        # act
        DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_diode_rqd_a, u_diode_rqd_a_name)

        # assert
        u_diode_rqd_a_processed = self.parsePmSignal('U_DIODE_RQD_A_Processed')

        self.assertListEqual(list(u_diode_rqd_a_processed[u_diode_rqd_a_name].values), list(u_diode_rqd_a[u_diode_rqd_a_name].values))

    def test_U_DIODE_RQD_B(self):
        # arrange
        u_diode_rqd_b = self.parsePmSignal('U_DIODE_RQD_B')
        u_diode_rqd_b_name = 'MQ.25R4:U_DIODE_RQD'
        timestamp_iqps_qh = '2018-12-09 09:55:47.668'
        timestamp_unix = Time.convertToUnixTimestamp(timestamp_iqps_qh)
        time_corr = 0.0

        # act
        DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_diode_rqd_b, u_diode_rqd_b_name)

        # assert
        u_deiode_rqd_b_processed = self.parsePmSignal('U_DIODE_RQD_B_Processed')

        self.assertListEqual(list(u_deiode_rqd_b_processed[u_diode_rqd_b_name].values), list(u_diode_rqd_b[u_diode_rqd_b_name].values))

    def test_U_REF_N1(self):
        # arrange
        u_ref_n1 = self.parsePmSignal('U_REF_N1')
        u_ref_n1_name = 'DQQDS.B26R4.RQD.A45:U_REF_N1'
        timestamp_iqps_qh = '2018-12-09 09:55:47.668'
        timestamp_unix = Time.convertToUnixTimestamp(timestamp_iqps_qh)
        time_corr = 0.0

        # act
        DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_ref_n1, u_ref_n1_name)

        # assert
        u_ref_n1_processed = self.parsePmSignal('U_REF_N1_Processed')

        self.assertListEqual(list(u_ref_n1_processed[u_ref_n1_name].values), list(u_ref_n1[u_ref_n1_name].values))

    def test_IAB_I_A(self):
        # arrange
        u_diode_rqd_a = self.parsePmSignal('U_DIODE_RQD_A')
        u_diode_rqd_a_name = 'MQ.25R4:U_DIODE_RQD'

        iab_i_a = self.parsePmSignal('IAB.I_A')
        iab_i_a_name = 'IAB.I_A'
        timestamp_iqps_qh = '2018-12-09 09:55:47.668'
        timestamp_unix = Time.convertToUnixTimestamp(timestamp_iqps_qh)
        time_corr = 0.0

        # act
        u_diode_rqd_a = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_diode_rqd_a, u_diode_rqd_a_name)

        index_start_diode = u_diode_rqd_a.index[0]
        index_end_diode = u_diode_rqd_a.index[-1]
        iab_i_a = DiodeLeadResistanceAnalysis.preprocess_pm_pc_signal(iab_i_a, iab_i_a_name, index_end_diode, index_start_diode, timestamp_unix)

        # assert
        iab_i_a_processed = self.parsePmSignal('IAB.I_A_Processed')

        self.assertListEqual(list(iab_i_a_processed[iab_i_a_name].values), list(iab_i_a[iab_i_a_name].values))

    def test_calculate_u_res_res_pm(self):
        # arrange
        u_diode_rqd_a = self.parsePmSignal('U_DIODE_RQD_A')
        u_diode_rqd_a_name = 'MQ.25R4:U_DIODE_RQD'

        u_diode_rqd_b = self.parsePmSignal('U_DIODE_RQD_B')
        u_diode_rqd_b_name = 'MQ.25R4:U_DIODE_RQD'

        iab_i_a = self.parsePmSignal('IAB.I_A')
        iab_i_a_name = 'IAB.I_A'

        timestamp_iqps_qh = '2018-12-09 09:55:47.668'
        timestamp_unix = Time.convertToUnixTimestamp(timestamp_iqps_qh)
        time_corr = 0.0

        # act
        u_diode_rqd_a = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_diode_rqd_a, u_diode_rqd_a_name)
        u_diode_rqd_a_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_diode_rqd_a)
        #
        u_diode_rqd_b = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_diode_rqd_b, u_diode_rqd_b_name)
        u_diode_rqd_b_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_diode_rqd_b)

        index_start_diode = u_diode_rqd_a.index[0]
        index_end_diode = u_diode_rqd_a.index[-1]
        iab_i_a = DiodeLeadResistanceAnalysis.preprocess_pm_pc_signal(iab_i_a, iab_i_a_name, index_end_diode, index_start_diode, timestamp_unix)
        iab_i_a_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(iab_i_a)

        #
        t_max = u_diode_rqd_a.idxmax(axis=0, skipna=True)[0]
        #
        # DataFrame operation
        uD = pd.DataFrame()
        uD['U_DIODE_A'] = u_diode_rqd_a_resampled[u_diode_rqd_a_name]
        uD['U_DIODE_B'] = u_diode_rqd_b_resampled[u_diode_rqd_b_name].values
        uD['IAB.I_A'] = iab_i_a_resampled[iab_i_a_name].values

        DiodeLeadResistanceAnalysis.calculate_resistence(t_max, uD)

        # assert
        u_res = self.parsePmSignal('U_RES')
        for u_d_min_exp, u_d_min_act in zip(list(u_res["U_RES"].values), list(uD['U_RES'].values)):
            self.assertAlmostEqual(u_d_min_exp, u_d_min_act, places=0)

        res = self.parsePmSignal('RES')
        for u_d_min_exp, u_d_min_act in zip(list(res["RES"].values), list(uD['RES'].values)):
            self.assertAlmostEqual(u_d_min_exp, u_d_min_act, places=5)

        res_max, t_res_max, i_res_max = DiodeLeadResistanceAnalysis.extract_res_max_time_current_pm(uD)
        self.assertEqual(32, res_max)
        self.assertAlmostEqual(0.471735280772, t_res_max, places=6)
        self.assertAlmostEqual(11157.375496, i_res_max, places=6)
