import os
import math
import unittest
import json
import pandas as pd
from pathlib import Path

from lhcsmapi.dbsignal.SignalElectrical import SignalElectrical
from lhcsmapi.metadata.Metadata import Metadata
from lhcsmapi.post_mortem.PmDbSignal import PmDbSignal
from lhcsmapi.analysis.QuenchHeaterAnalysis import QuenchHeaterAnalysis

class TestQuenchHeaterAnalysis(unittest.TestCase):

    @staticmethod
    def parsePmSignal(eventTime, source, signalName):
        system = "QPS"
        className = "DQAMCNMB_PMHSU"

        pms = PmDbSignal(system=system, className=className, source=source, signalName=signalName, eventTime=eventTime)
        path = Path(os.path.dirname(__file__))
        filePath = os.path.join(str(path.parent), 'resources/quench_heater_analysis/{}_{}_{}.json'.format(eventTime, source, signalName))
        with open(filePath) as json_file:
            response = json.load(json_file)
            content = response['content'][0]
            nameOfValue = pms._getNamesOfValue(content)
            pms._getTimeValue(nameOfValue, content)
            return pms._getSignalDataframe(signalName)

    @classmethod
    def setUpClass(self):
        circuitType = 'RB'
        systemType = 'QPS'
        component = 'QH'
        magnetTypeToSignalMetadata = Metadata.getMetadata(circuitType, systemType)
        qpsMetadata = magnetTypeToSignalMetadata[systemType]
        uNames = qpsMetadata[component]['PM']['U_HDS']
        iNames = qpsMetadata[component]['PM']['I_HDS']
        rNames = qpsMetadata[component]['R_HDS']
        source = "A10L7"

        # Parse selected discharge
        eventTime = 1397227077614000000

        dfList = []
        for signalName in uNames + iNames:
            dfList.append(TestQuenchHeaterAnalysis.parsePmSignal(eventTime, source, signalName))

        self.df = pd.concat(dfList, axis=1)

        self.dfSync = QuenchHeaterAnalysis.preProcessDischarge(self.df)

        # Calculate resistance
        if circuitType == 'RB':
            self.dfSync = SignalElectrical.calculateResistances(self.dfSync, uNames, iNames, rNames)

        self.features = QuenchHeaterAnalysis.extractFeatures(self.dfSync, eventTime, uNames, iNames, rNames)

        # Parse reference discharge
        eventTimeRef = 1412844934913000000

        dfRefList = []
        for signalName in uNames + iNames:
            dfRefList.append(TestQuenchHeaterAnalysis.parsePmSignal(eventTimeRef, source, signalName))

        self.dfRef = pd.concat(dfRefList, axis=1)

        self.dfRefSync = QuenchHeaterAnalysis.preProcessDischarge(self.dfRef)

        if circuitType == 'RB':
            self.dfRefSync = SignalElectrical.calculateResistances(self.dfRefSync, uNames, iNames, rNames)

        self.featuresRef = QuenchHeaterAnalysis.extractFeatures(self.dfRefSync, eventTimeRef, uNames, iNames, rNames)

        self.c2c = QuenchHeaterAnalysis.performCurveToCurveComparison(self.dfSync, self.dfRefSync, self.featuresRef, iNames, rNames, uNames)

    def test_SelectedSignal_count(self):
        self.assertEqual(16384, self.df['U_HDS_1'].count())
        self.assertEqual(16384, self.df['U_HDS_2'].count())
        self.assertEqual(16384, self.df['U_HDS_3'].count())
        self.assertEqual(16384, self.df['U_HDS_4'].count())
        self.assertEqual(16384, self.df['I_HDS_1'].count())
        self.assertEqual(16384, self.df['I_HDS_2'].count())
        self.assertEqual(16384, self.df['I_HDS_3'].count())
        self.assertEqual(16384, self.df['I_HDS_4'].count())

    def test_SelectedSignal_min(self):
        self.assertAlmostEqual(0.0, self.df['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_4'].min(), places=5)

    def test_SelectedSignal_max(self):
        self.assertAlmostEqual(629.183100, self.df['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.df['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.df['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.df['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_4'].max(), places=5)

    def test_SelectedSignal_mean(self):
        self.assertAlmostEqual(1.609766, self.df['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(2.375004, self.df['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(5.300451, self.df['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(4.584578, self.df['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(0.316847, self.df['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(0.293024, self.df['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(0.304800, self.df['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(0.193104, self.df['I_HDS_4'].mean(), places=5)

    def test_SelectedSignal_std(self):
        self.assertAlmostEqual(8.572087, self.df['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(32.265665, self.df['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(67.029452, self.df['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(61.729987, self.df['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(4.576926, self.df['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(4.237029, self.df['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(4.451355, self.df['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(2.167571, self.df['I_HDS_4'].std(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_count(self):
        self.assertEqual(61207, self.dfSync['U_HDS_1'].count())
        self.assertEqual(61207, self.dfSync['U_HDS_2'].count())
        self.assertEqual(61207, self.dfSync['U_HDS_3'].count())
        self.assertEqual(61207, self.dfSync['U_HDS_4'].count())
        self.assertEqual(61207, self.dfSync['I_HDS_1'].count())
        self.assertEqual(61207, self.dfSync['I_HDS_2'].count())
        self.assertEqual(61207, self.dfSync['I_HDS_3'].count())
        self.assertEqual(61207, self.dfSync['I_HDS_4'].count())

    def test_SelectedSignal_SyncDecayResampledSec_min(self):
        self.assertAlmostEqual(0.05749815, self.dfSync['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.05749815, self.dfSync['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.05749815, self.dfSync['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.05749815, self.dfSync['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_4'].min(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_max(self):
        self.assertAlmostEqual(98.187675, self.dfSync['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.dfSync['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.dfSync['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.dfSync['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.dfSync['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.dfSync['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.dfSync['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.dfSync['I_HDS_4'].max(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_mean(self):
        self.assertAlmostEqual(0.8415, self.dfSync['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(0.97911, self.dfSync['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(1.25714, self.dfSync['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(1.28054, self.dfSync['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(0.11405, self.dfSync['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(0.1051, self.dfSync['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(0.11076, self.dfSync['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(0.08633, self.dfSync['I_HDS_4'].mean(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_std(self):
        self.assertAlmostEqual(1.70862, self.dfSync['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(16.1227, self.dfSync['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(23.24638, self.dfSync['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(23.24598, self.dfSync['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(2.31367, self.dfSync['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(2.1449, self.dfSync['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(2.31372, self.dfSync['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(1.03655, self.dfSync['I_HDS_4'].std(), places=5)

    def test_SelectedSignal_CTimeChargeApproach(self):
        self.assertAlmostEqual(0.00326, self.features['U_HDS_1_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00315, self.features['U_HDS_2_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00032, self.features['U_HDS_3_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00032, self.features['U_HDS_4_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00029, self.features['I_HDS_1_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00248, self.features['I_HDS_2_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00028, self.features['I_HDS_3_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00303, self.features['I_HDS_4_chargeApproach'].values[0], places=5)

    def test_SelectedSignal_CTimeEnergyApproach(self):
        self.assertAlmostEqual(0.00033, self.features['U_HDS_1_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.01667, self.features['U_HDS_2_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['U_HDS_3_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['U_HDS_4_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['I_HDS_1_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.01597, self.features['I_HDS_2_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['I_HDS_3_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.00811, self.features['I_HDS_4_energyApproach'].values[0], places=5)

    def test_SelectedSignal_CTimeLinLog(self):
        self.assertAlmostEqual(-0.51955, self.features['U_HDS_1_linLog'].values[0], places=5)
        self.assertAlmostEqual(-0.53831, self.features['U_HDS_2_linLog'].values[0], places=5)
        self.assertAlmostEqual(-0.52061, self.features['U_HDS_3_linLog'].values[0], places=5)
        self.assertAlmostEqual(-0.51535, self.features['U_HDS_4_linLog'].values[0], places=5)
        self.assertAlmostEqual(787.40157, self.features['I_HDS_1_linLog'].values[0], places=5)
        self.assertAlmostEqual(787.40157, self.features['I_HDS_2_linLog'].values[0], places=5)
        self.assertAlmostEqual(787.40157, self.features['I_HDS_3_linLog'].values[0], places=5)
        self.assertAlmostEqual(787.40157, self.features['I_HDS_4_linLog'].values[0], places=5)

    def test_SelectedSignal_CTime_std(self):
        self.assertTrue(math.isnan(self.features['U_HDS_1_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['U_HDS_2_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['U_HDS_3_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['U_HDS_4_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_1_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_2_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_3_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_4_cTime_std'].values[0]))

    def test_SelectedSignal_CTime_mean(self):
        self.assertTrue(math.isinf(self.features['U_HDS_1_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['U_HDS_2_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['U_HDS_3_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['U_HDS_4_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_1_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_2_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_3_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_4_cTime_mean'].values[0]))

    def test_SelectedSignal_Initial(self):
        self.assertAlmostEqual(89.19197, self.features['U_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(597.37928, self.features['U_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(1256.0471, self.features['U_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(1256.0471, self.features['U_HDS_4_initial'].values[0], places=5)
        self.assertAlmostEqual(125.00146, self.features['I_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(107.71716, self.features['I_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(125.00146, self.features['I_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(24.34814, self.features['I_HDS_4_initial'].values[0], places=5)

    def test_SelectedSignal_Final(self):
        self.assertAlmostEqual(0.84331, self.features['U_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(0.78581, self.features['U_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(0.84671, self.features['U_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(0.88164, self.features['U_HDS_4_final'].values[0], places=5)
        self.assertAlmostEqual(0.07439, self.features['I_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(0.07057, self.features['I_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(0.07064, self.features['I_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(0.0782, self.features['I_HDS_4_final'].values[0], places=5)

    def test_SelectedSignal_InitialResistance(self):
        self.assertAlmostEqual(0.71353, self.features['R_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(5.54581, self.features['R_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(10.04826, self.features['R_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(51.58699, self.features['R_HDS_4_initial'].values[0], places=5)

    def test_SelectedSignal_SimilarityMatrix(self):
        self.assertAlmostEqual(2.92628, self.features['U_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(4.25487, self.features['U_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(4.25488, self.features['U_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(3.00473, self.features['U_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(3.00475, self.features['U_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(0.00501, self.features['U_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(1.51699, self.features['I_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(0.00699, self.features['I_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(3.71682, self.features['I_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(1.51697, self.features['I_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(3.35433, self.features['I_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(3.71683, self.features['I_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_11'].values[0], places=5)

    def test_ReferenceSignal_count(self):
        self.assertEqual(16384, self.dfRef['U_HDS_1'].count())
        self.assertEqual(16384, self.dfRef['U_HDS_2'].count())
        self.assertEqual(16384, self.dfRef['U_HDS_3'].count())
        self.assertEqual(16384, self.dfRef['U_HDS_4'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_1'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_2'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_3'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_4'].count())

    def test_ReferenceSignal_min(self):
        self.assertAlmostEqual(36.281334, self.dfRef['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(28.423252, self.dfRef['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(34.709717, self.dfRef['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(37.661290, self.dfRef['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(0.009537, self.dfRef['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(1.970344, self.dfRef['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfRef['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.022889, self.dfRef['I_HDS_4'].min(), places=5)

    def test_ReferenceSignal_max(self):
        self.assertAlmostEqual(989.639040, self.dfRef['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(935.034900, self.dfRef['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(949.352000, self.dfRef['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(933.827450, self.dfRef['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(86.859180, self.dfRef['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(88.388916, self.dfRef['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(83.870285, self.dfRef['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(88.268745, self.dfRef['I_HDS_4'].max(), places=5)

    def test_ReferenceSignal_mean(self):
        self.assertAlmostEqual(441.296557, self.dfRef['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(421.297825, self.dfRef['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(444.610057, self.dfRef['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(440.431096, self.dfRef['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(34.979697, self.dfRef['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(33.770704, self.dfRef['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(34.113873, self.dfRef['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(34.536778, self.dfRef['I_HDS_4'].mean(), places=5)

    def test_ReferenceSignal_std(self):
        self.assertAlmostEqual(286.160681, self.dfRef['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(284.258130, self.dfRef['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(288.792182, self.dfRef['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(282.673923, self.dfRef['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(23.721566, self.dfRef['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(23.603027, self.dfRef['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(23.316205, self.dfRef['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(23.233123, self.dfRef['I_HDS_4'].std(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_count(self):
        self.assertEqual(61283, self.dfRefSync['U_HDS_1'].count())
        self.assertEqual(61283, self.dfRefSync['U_HDS_2'].count())
        self.assertEqual(61283, self.dfRefSync['U_HDS_3'].count())
        self.assertEqual(61283, self.dfRefSync['U_HDS_4'].count())
        self.assertEqual(61283, self.dfRefSync['I_HDS_1'].count())
        self.assertEqual(61283, self.dfRefSync['I_HDS_2'].count())
        self.assertEqual(61283, self.dfRefSync['I_HDS_3'].count())
        self.assertEqual(61283, self.dfRefSync['I_HDS_4'].count())

    def test_ReferenceSignal_SyncDecayResampledSec_min(self):
        self.assertAlmostEqual(36.28133, self.dfRefSync['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(28.44242, self.dfRefSync['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(34.72888, self.dfRefSync['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(37.66129, self.dfRefSync['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(2.613138, self.dfRefSync['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(1.970344, self.dfRefSync['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(2.401416, self.dfRefSync['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(2.693249, self.dfRefSync['I_HDS_4'].min(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_max(self):
        self.assertAlmostEqual(930.7226, self.dfRefSync['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(912.5723, self.dfRefSync['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(935.2458, self.dfRefSync['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(921.81036, self.dfRefSync['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(81.18657, self.dfRefSync['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(79.20669, self.dfRefSync['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(79.92197, self.dfRefSync['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(79.88763, self.dfRefSync['I_HDS_4'].max(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_mean(self):
        self.assertAlmostEqual(252.05218, self.dfRefSync['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(234.11997, self.dfRefSync['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(253.33055, self.dfRefSync['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(253.22555, self.dfRefSync['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(19.81358, self.dfRefSync['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(18.44679, self.dfRefSync['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(19.42279, self.dfRefSync['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(19.73928, self.dfRefSync['I_HDS_4'].mean(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_std(self):
        self.assertAlmostEqual(228.20034, self.dfRefSync['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(224.73858, self.dfRefSync['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(231.01783, self.dfRefSync['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(226.1438, self.dfRefSync['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(18.46009, self.dfRefSync['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(18.2011, self.dfRefSync['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(18.24874, self.dfRefSync['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(18.14661, self.dfRefSync['I_HDS_4'].std(), places=5)

    def test_ReferenceSignal_CTimeChargeApproach(self):
        self.assertAlmostEqual(0.08994, self.featuresRef['U_HDS_1_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.08452, self.featuresRef['U_HDS_2_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.08979, self.featuresRef['U_HDS_3_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.09141, self.featuresRef['U_HDS_4_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.08068, self.featuresRef['I_HDS_1_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.07623, self.featuresRef['I_HDS_2_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.07997, self.featuresRef['I_HDS_3_chargeApproach'].values[0], places=5)
        self.assertAlmostEqual(0.08237, self.featuresRef['I_HDS_4_chargeApproach'].values[0], places=5)

    def test_ReferenceSignal_CTimeEnergyApproach(self):
        self.assertAlmostEqual(0.08532, self.featuresRef['U_HDS_1_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.0808, self.featuresRef['U_HDS_2_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.0859, self.featuresRef['U_HDS_3_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.08673, self.featuresRef['U_HDS_4_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.07142, self.featuresRef['I_HDS_1_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.06837, self.featuresRef['I_HDS_2_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.07104, self.featuresRef['I_HDS_3_energyApproach'].values[0], places=5)
        self.assertAlmostEqual(0.07329, self.featuresRef['I_HDS_4_energyApproach'].values[0], places=5)

    def test_ReferenceSignal_CTimeLinLog(self):
        self.assertAlmostEqual(0.09894, self.featuresRef['U_HDS_1_linLog'].values[0], places=5)
        self.assertAlmostEqual(0.09208, self.featuresRef['U_HDS_2_linLog'].values[0], places=5)
        self.assertAlmostEqual(0.09722, self.featuresRef['U_HDS_3_linLog'].values[0], places=5)
        self.assertAlmostEqual(0.10043, self.featuresRef['U_HDS_4_linLog'].values[0], places=5)
        self.assertAlmostEqual(0.09691, self.featuresRef['I_HDS_1_linLog'].values[0], places=5)
        self.assertAlmostEqual(0.08956, self.featuresRef['I_HDS_2_linLog'].values[0], places=5)
        self.assertAlmostEqual(0.0949, self.featuresRef['I_HDS_3_linLog'].values[0], places=5)
        self.assertAlmostEqual(0.09835, self.featuresRef['I_HDS_4_linLog'].values[0], places=5)

    def test_ReferenceSignal_CTime_std(self):
        self.assertAlmostEqual(0.01163, self.featuresRef['U_HDS_1_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.01055, self.featuresRef['U_HDS_2_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.00983, self.featuresRef['U_HDS_3_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.01169, self.featuresRef['U_HDS_4_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.01252, self.featuresRef['I_HDS_1_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.00982, self.featuresRef['I_HDS_2_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.01094, self.featuresRef['I_HDS_3_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.0127, self.featuresRef['I_HDS_4_cTime_std'].values[0], places=5)

    def test_ReferenceSignal_CTime_mean(self):
        self.assertAlmostEqual(0.11167, self.featuresRef['U_HDS_1_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.10413, self.featuresRef['U_HDS_2_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.10963, self.featuresRef['U_HDS_3_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.11335, self.featuresRef['U_HDS_4_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.10682, self.featuresRef['I_HDS_1_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.09853, self.featuresRef['I_HDS_2_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.10434, self.featuresRef['I_HDS_3_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.10842, self.featuresRef['I_HDS_4_cTime_mean'].values[0], places=5)

    def test_ReferenceSignal_Initial(self):
        self.assertAlmostEqual(930.19195, self.featuresRef['U_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(912.03464, self.featuresRef['U_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(934.69296, self.featuresRef['U_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(921.27572, self.featuresRef['U_HDS_4_initial'].values[0], places=5)
        self.assertAlmostEqual(81.01189, self.featuresRef['I_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(79.11433, self.featuresRef['I_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(79.48016, self.featuresRef['I_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(79.48707, self.featuresRef['I_HDS_4_initial'].values[0], places=5)

    def test_ReferenceSignal_Final(self):
        self.assertAlmostEqual(36.2976, self.featuresRef['U_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(28.45868, self.featuresRef['U_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(34.73645, self.featuresRef['U_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(37.68449, self.featuresRef['U_HDS_4_final'].values[0], places=5)
        self.assertAlmostEqual(2.61389, self.featuresRef['I_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(1.97196, self.featuresRef['I_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(2.40297, self.featuresRef['I_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(2.69556, self.featuresRef['I_HDS_4_final'].values[0], places=5)

    def test_ReferenceSignal_InitialResistance(self):
        self.assertAlmostEqual(11.48217, self.featuresRef['R_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(11.52806, self.featuresRef['R_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(11.76008, self.featuresRef['R_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(11.59026, self.featuresRef['R_HDS_4_initial'].values[0], places=5)

    def test_ReferenceSignal_SimilarityMatrix(self):
        self.assertAlmostEqual(5.10669, self.featuresRef['U_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(0.89703, self.featuresRef['U_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(0.72483, self.featuresRef['U_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(5.67548, self.featuresRef['U_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(5.41926, self.featuresRef['U_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(1.34627, self.featuresRef['U_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(4.42103, self.featuresRef['I_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(1.41063, self.featuresRef['I_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(1.04858, self.featuresRef['I_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(3.20083, self.featuresRef['I_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(4.18134, self.featuresRef['I_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(1.05166, self.featuresRef['I_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(12.96017, self.featuresRef['R_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(30.39083, self.featuresRef['R_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(8.82041, self.featuresRef['R_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(26.90587, self.featuresRef['R_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(13.41515, self.featuresRef['R_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(21.81435, self.featuresRef['R_HDS_SimilarityMatrix_11'].values[0], places=5)

    def test_curve2curveComparison(self):
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_1_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_2_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_3_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_4_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_1_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_2_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_3_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_4_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.4277, self.c2c['R_HDS_1_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.56301, self.c2c['R_HDS_2_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.95393, self.c2c['R_HDS_3_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.31081, self.c2c['R_HDS_4_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)