import unittest
from unittest.mock import patch
import requests
from lhcsmapi.post_mortem.PmDbSignal import PmDbSignal as SignalPM
import pandas as pd
from datetime import datetime

class TestPmDbSignal(unittest.TestCase):

    def setUp(self):
        self.SYSTEM = "QPS"
        self.CLASS_NAME = "DQAMCNMB_PMHSU"
        self.SOURCE = "B20L5"
        self.SIGNAL_NAME = "I_HDS_4"
        self.TIMESTAMP = 1426220469491000000
        self.defaultSignalPM = SignalPM()

    def tearDown(self):
        pass

    def assertEqualSignals(self, firstSignalPM, secondSignalPM):
        """Assert each property of firstSignalPM against of the same property of secondSignalPM"""
        self.assertEqual(firstSignalPM.System, secondSignalPM.System)
        self.assertEqual(firstSignalPM.ClassName, secondSignalPM.ClassName)
        self.assertEqual(firstSignalPM.Source, secondSignalPM.Source)
        self.assertEqual(firstSignalPM.SignalName, secondSignalPM.SignalName)
        self.assertEqual(firstSignalPM.EventTime, secondSignalPM.EventTime)
        self.assertEqual(firstSignalPM.StartTime, secondSignalPM.StartTime)
        self.assertEqual(firstSignalPM.IntervTime, secondSignalPM.StartTime)
        self.assertEqual(firstSignalPM.Unit, secondSignalPM.Unit)
        self.assertEqual(firstSignalPM.Value, secondSignalPM.Value)
        self.assertEqual(firstSignalPM.Time, secondSignalPM.Time)

    def test_toString(self):
        # arrange
        pms = SignalPM().buildSignal(system=self.SYSTEM, className=self.CLASS_NAME, source=self.SOURCE,
                              signalName=self.SIGNAL_NAME, eventTime=self.TIMESTAMP)
        # act
        pmsStringAct = str(pms)

        # assert
        pmsStringExp = 'Db: PM, Signal: I_HDS_4, eventTime = 1426220469491000000 (2015-03-13 05:21:09.491000+01:00), system=QPS, source=B20L5, className=DQAMCNMB_PMHSU'
        self.assertEqual(pmsStringExp, pmsStringAct)

    def test_createSignalPM_noParameters_defaultSignalCreated(self):
        """Tests that the default CalsSignal is created if no arguments are passed to the createSignalPM method"""

        # act:
        pmSignal = SignalPM.buildSignal()

        # assert:
        self.assertEqualSignals(pmSignal, self.defaultSignalPM)

    def test_createSignalPM_initParametersProvided_CorrectSignalCreated(self):
        """Tests that the correct CalsSignal is created if arbitrary arguments are passed to the createSignalPM method"""

        # arrange:
        now = datetime.now()
        arbitraryPMSignal = SignalPM("System", "ClassName", "source", "test", pd.Timestamp(now))

        # act:
        pmSignal = SignalPM.buildSignal(system="System",
                                        className="ClassName",
                                        source="source",
                                        signalName="test",
                                        eventTime=pd.Timestamp(now))

        # assert:
        self.assertEqualSignals(pmSignal, arbitraryPMSignal)

    def test_createSignalPM_parameterIsInvalidType_TypeErrorIsThrown(self):
        """Tests that when kwargs contain a username keyword that isn't a string, a TypeError
        is thrown"""

        now = datetime.now()

        with self.assertRaises(TypeError):
            SignalPM.buildSignal(system=1000,
                                 className="ClassName",
                                 source="source",
                                 signalName="test",
                                 eventTime=pd.Timestamp(now))


    # needs read and getTimeValue tests


    def test_getNamesOfValues_correctDBAccessWriteCalled(self):
        """Test that write correctly call the write method of the correct DbSignal type dependent on the api parameter"""

        # arrange:
        uut = SignalPM()

        testCases = [("zero", {'namesAndValues': []}, []),
                     ('one', {'namesAndValues': [{'name': 'name1'}]}, ['name1']),
                     ('many', {'namesAndValues': [{'name': 'name1'}, {'name': 'name2'}]}, ['name1', 'name2'])]

        for testName, content, valueNames in testCases:
            with self.subTest(name=testName):

                # act:
                result = uut._getNamesOfValue(content)

                # assert:
                self.assertEqual(result, valueNames)

    def test_parseSignal_queryingFailed_RaisesError(self):

        pmSignal = SignalPM("System", "ClassName", "source", "test", pd.Timestamp(datetime.now()))

        with patch.object(requests, 'get') as mock_get:
            mock_get.side_effect = Exception()

            with self.assertRaises(Exception):
                pmSignal.parseSignal()
